# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""This module is for seasonality detection.

We provide two seasonality detector: ACFDetector and FFTDetector. ACFDetector uses
autocorrelation function to find seasonality, while FFTDetector uses Fast Fourier
Transform to detect seasonality.

Typical usage example:

>>> timeseries = TimeSeriesData(...)
>>> # initialize detector
>>> detector = ACFDetector(timeseries)
>>> # run detector
>>> detector.detector(diff=1, alpha = 0.01)
>>> # seasonality decomposition, returns trend, seasonal, residual term
>>> detector.remover()
>>> # plot acf and decompsition results
>>> detector.plot()
"""

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.fftpack as fp
import statsmodels.api as sm
from scipy.signal import find_peaks  # @manual
from statsmodels.tsa.stattools import acf

from kats.consts import TimeSeriesData
from kats.detectors.detector import Detector
from kats.utils.decomposition import TimeSeriesDecomposition

# from numpy.typing import ArrayLike
ArrayLike = Union[np.ndarray, Sequence[float]]


class ACFDetector(Detector):
    """Autocorrelation function seasonality detector.

    Use acf to detect seasonality, and find out the potential cycle lengths

    Attributes:
        data: The input time series data from TimeSeriesData
        decomposed: A bool indicate if we decomposed the time series into trend,
            seasonal and residual.
    """

    ts_diff: Optional[ArrayLike] = None
    lags: Optional[int] = None
    seasonality: Optional[List[int]] = None
    seasonality_detected: bool = False
    decompose: Optional[TimeSeriesDecomposition] = None

    def __init__(self, data: TimeSeriesData) -> None:
        super().__init__(data=data)
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)
        self.decomposed = False

    def _get_seasonality_length(self, d: List[int]) -> List[int]:
        out = []
        while d:
            k = d.pop(0)
            d = [i for i in d if i % k != 0]
            out.append(k)
        return out

    # pyre-fixme[14]: Inconsistent override [14]: `kats.detectors.seasonality.ACFDetector.detector` overrides method defined in `Detector` inconsistently. Could not find parameter `method` in overriding signature.
    # pyre-fixme[14]: `kats.detectors.seasonality.ACFDetector.detector` overrides method defined in `Detector` inconsistently. Returned type `Dict[str, typing.Any]` is not a subtype of the overridden return `None`.
    # pyre-fixme[15]: `detector` overrides method defined in `Detector` inconsistently.
    def detector(self, lags: Optional[int] = None, diff: int = 1, alpha: Optional[float] = 0.01) -> Dict[str, Any]:
        """Detect seasonality

        This method runs acf and returns if seasonality detected in the given time series
        and potential cycle lengths

        Args:
            lags: Optional; int; the maximum lags we used in acf.
            diff: Optional; int; times of diff run on timeseries to remove trend before
                apply acf.
            alpha: Optional; float; significant level we use the calcualte
                autocorrelation confidence interval.

        Returns:
            A dict contains
                - seasonality_presence: bool, if seasonality detected
                - seasonalities: List[int], potential seasonlities cycle length(s)
        """

        # Use array to store the data
        ts = self.data.value.values
        ts_diff = ts
        for _ in range(diff):
            ts_diff = np.diff(ts_diff)
        self.ts_diff = ts_diff

        if lags is None:
            lags = int(len(ts) / 3)
        self.lags = lags
        ac, confint, qstat, qval = acf(ts_diff, nlags=lags, qstat=True, alpha=alpha)
        # get seasonality cycle length
        raw_seasonality = []
        for i, _int in enumerate(confint):
            if _int[0] >= 0 and i > 1:
                raw_seasonality.append(i)
        self.seasonality = self._get_seasonality_length(raw_seasonality)
        self.seasonality_detected = True if self.seasonality else False

        return {
            "seasonality_presence": self.seasonality_detected,
            "seasonalities": self.seasonality,
        }

    def plot(self, **kwargs: Any) -> Union[plt.Axes, Sequence[plt.Axes]]:
        """Plot detection results.

        Args:
            None

        Returns:
            The matplotlib Axes.
        """
        fig = sm.graphics.tsa.plot_acf(self.ts_diff, lags=self.lags)
        ax1 = fig.gca()
        if self.decomposed:
            decompose = self.decompose
            assert decompose is not None
            axs = decompose.plot()
            return (ax1, *axs)
        return ax1

    # pyre-fixme[14]: `kats.detectors.seasonality.ACFDetector.remover` overrides method defined in `Detector` inconsistently. Could not find parameter `interpolate` in overriding signature.
    # pyre-fixme[15]: `kats.detectors.seasonality.ACFDetector.remover` overrides method defined in `Detector` inconsistently. Returned type `Optional[Dict[str, typing.Any]]` is not a subtype of the overridden return `TimeSeriesData`.
    def remover(
        self,
        decom: Type[TimeSeriesDecomposition] = TimeSeriesDecomposition,
        model: str = "additive",
        decompose_any_way: bool = False,
    ) -> Optional[Dict[str, Any]]:
        """Remove the seasonality in the time series

        Args:
            decom: Optional; decomposition method.
            model: Optional; model used for decomposition.
            decompose_any_way: Optional; bool; decompose the time series even when
                seasonality is not detected in the time series.

        Returns:
            decomposition results of the decomposition method.
        """

        if decompose_any_way or self.seasonality_detected:
            self.decompose = decompose = decom(self.data, model)
            result = decompose.decomposer()
            self.decomposed = True
            return result
        else:
            logging.info("No seasonality detected, not running decomposition")


class FFTDetector(Detector):
    """Fast Fourier Transform Seasoanlity detector

    Use Fast Fourier Transform to detect seasonality, and find out the
    potential cycle's length.

    Attributes:
        data: The input time series data from TimeSeriesData.
    """

    def __init__(self, data: TimeSeriesData) -> None:
        super().__init__(data=data)
        if not self.data.is_univariate():
            msg = "The provided time series data is not univariate."
            logging.error(msg)
            raise ValueError(msg)

    # pyre-fixme[14]: `detector` overrides method defined in `Detector` inconsistently.
    # pyre-fixme[15]: `detector` overrides method defined in `Detector` inconsistently.
    def detector(self, sample_spacing: float = 1.0, mad_threshold: float = 6.0) -> Dict[str, Any]:
        """Detect seasonality with FFT

        Args:
            sample_spacing: Optional; float; scaling FFT for a different time unit.
                I.e. for hourly time series, sample_spacing=24.0, FFT x axis will be
                1/day.
            mad_threshold: Optional; float; constant for the outlier algorithm for peak
                detector. The larger the value the less sensitive the outlier algorithm
                is.

        Returns:
            FFT Plot with peaks, selected peaks, and outlier boundary line.
        """

        fft = self.get_fft(sample_spacing)
        _, orig_peaks, peaks = self.get_fft_peaks(fft, mad_threshold)
        seasonality_presence = len(peaks.index) > 0
        selected_seasonalities = []
        if seasonality_presence:
            selected_seasonalities = peaks["freq"].transform(lambda x: 1 / x).tolist()

        return {
            "seasonality_presence": seasonality_presence,
            "seasonalities": selected_seasonalities,
        }

    def get_fft(self, sample_spacing: float = 1.0) -> pd.DataFrame:
        """Computes FFT

        Args:
            sample_spacing: Optional; scaling FFT for a different time unit.
                I.e. for hourly time series, sample_spacing=24.0 FFT x axis will be 1/day.

        Returns:
            DataFrame with columns 'freq' and 'ampl'.
        """
        data_fft = fp.fft(self.data.value.values)
        data_psd = np.abs(data_fft) ** 2
        fftfreq = fp.fftfreq(len(data_psd), 1.0 / sample_spacing)
        pos_freq_ix = fftfreq > 0

        freq = (fftfreq[pos_freq_ix],)
        ampl = (10 * np.log10(data_psd[pos_freq_ix]),)

        return pd.DataFrame({"freq": freq[0], "ampl": ampl[0]}, copy=False)

    def get_fft_peaks(self, fft: pd.DataFrame, mad_threshold: float = 6.0) -> Tuple[float, pd.DataFrame, pd.DataFrame]:
        """Computes peaks in fft, selects the highest peaks (outliers) and
            removes the harmonics (multiplies of the base harmonics found)

        Args:
            fft: FFT computed by FFTDetector.get_fft
            sample_spacing: Optional; scaling FFT for a different time unit.
                I.e. for hourly time series, sample_spacing=24.0 FFT x axis will be 1/day.
            mad_threshold: Optional; constant for the outlier algorithm for peak detector.
                The larger the value the less sensitive the outlier algorithm is.

        Returns:
            outlier threshold, peaks, selected peaks.
        """
        pos_fft = fft.loc[fft["ampl"] > 0]
        median = pos_fft["ampl"].median()
        pos_fft_above_med = pos_fft[pos_fft["ampl"] > median]
        mad = pos_fft_above_med["ampl"].mad()

        threshold = median + mad * mad_threshold

        peak_indices = find_peaks(fft["ampl"], threshold=0.1)
        peaks = fft.loc[peak_indices[0], :]

        orig_peaks = peaks.copy()

        peaks = peaks.loc[peaks["ampl"] > threshold].copy()
        peaks["Remove"] = [False] * len(peaks.index)
        peaks.reset_index(inplace=True)

        # Filter out harmonics
        for idx1 in range(len(peaks)):
            curr = peaks.loc[idx1, "freq"]
            for idx2 in range(idx1 + 1, len(peaks)):
                if peaks.loc[idx2, "Remove"] is True:
                    continue
                fraction = (peaks.loc[idx2, "freq"] / curr) % 1
                if fraction < 0.01 or fraction > 0.99:
                    peaks.loc[idx2, "Remove"] = True
        peaks = peaks.loc[~peaks["Remove"]]
        peaks.drop(inplace=True, columns="Remove")
        return threshold, orig_peaks, peaks



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from . import compat  # noqa # usort: skip
from . import consts  # noqa # usort: skip
from . import data  # noqa # usort: skip
from . import detectors  # noqa # usort: skip
from . import metrics  # noqa # usort: skip
from . import models  # noqa # usort: skip
from . import tsfeatures  # noqa # usort: skip
from . import utils  # noqa # usort: skip



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""
This module implements the univariate Outlier Detection algorithm as a Detector Model.
"""
import json
from typing import Any
from typing import Optional

import numpy as np
import pandas as pd

from kats.consts import TimeSeriesData
from kats.detectors.detector import DetectorModel
from kats.detectors.detector_consts import AnomalyResponse
from kats.detectors.detector_consts import ConfidenceBand
from kats.detectors.outlier import OutlierDetector


class OutlierDetectorModel(DetectorModel):
    """Anamoly detection model based on outlier detector.

    A Detector Model detects outliers for a single time series by decomposing the time series
    and detecting outliers when the values in the residual time series are beyond the
    specified multiplier times the inter quartile range.

    Attributes:
        decomp: 'additive' or 'multiplicative' decomposition
        iqr_mult: iqr_mult * inter quartile range is used to classify outliers
        serialized_model: Optional; json, representing data from a previously serialized model.
    """

    decomp: str
    iqr_mult: float
    model: Optional[OutlierDetector] = None

    def __init__(
        self,
        decomp: str = "additive",
        iqr_mult: float = 3.0,
        serialized_model: Optional[bytes] = None,
    ) -> None:
        if serialized_model:
            model_dict = json.loads(serialized_model)
            self.decomp = model_dict["decomp"]
            self.iqr_mult = model_dict["iqr_mult"]
        else:
            self.decomp = decomp
            self.iqr_mult = iqr_mult

    def serialize(self) -> bytes:
        """Serialize the model into a json.

        Serialize the model into a json so it can be loaded later.

        Returns:
            json containing information of the model.
        """
        model_dict = {"decomp": self.decomp, "iqr_mult": self.iqr_mult}
        return json.dumps(model_dict).encode("utf-8")

    def fit(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> None:
        """Fit OutlierDetector.

        Fit OutlierDetector using both data and historical_data (if provided).

        Args:
            data: TimeSeriesData on which detection is run.
            historical_data: Optional; TimeSeriesData corresponding to history. History ends
                exactly where the data begins.

        Returns:
            None.
        """
        if historical_data is None:
            total_data = data
        else:
            historical_data.extend(data)
            total_data = historical_data

        self.model = OutlierDetector(data=total_data, decomp=self.decomp, iqr_mult=self.iqr_mult)
        self.model.detector()

    def predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """Get anomaly scores.

        Get anomaly scores only for data.

        Args:
            data: TimeSeriesData on which detection is run
            historical_data: Optional; TimeSeriesData corresponding to history. History ends
                exactly where the data begins.

        Returns:
            AnomalyResponse object. The length of this obj.ect is same as data. The score property
            gives the score for anomaly.
        """
        # When no iterpolate argument is given by default it is taking False
        if "interpolate" not in kwargs:
            interpolate = bool(False)
        else:
            interpolate = bool(kwargs["interpolate"])

        if self.model is None:
            self.fit(data=data, historical_data=historical_data)

        assert self.model is not None
        output_scores_df = self.model.output_scores
        output_detector_remover = self.model.remover(interpolate=interpolate)

        assert output_scores_df is not None
        output_scores_df = output_scores_df[output_scores_df.index >= data.time.min()]

        zeros = pd.DataFrame(np.zeros(shape=output_scores_df.shape), copy=False)
        # all fields other than scores and predicted_ts are left as TimeSeriesData with all zero values
        response = AnomalyResponse(
            scores=TimeSeriesData(
                time=data.time,
                value=output_scores_df,
            ),
            confidence_band=ConfidenceBand(
                upper=TimeSeriesData(time=data.time, value=zeros),
                lower=TimeSeriesData(time=data.time, value=zeros),
            ),
            predicted_ts=TimeSeriesData(
                time=output_detector_remover.time,
                value=pd.DataFrame(output_detector_remover.value, copy=False),
            ),
            anomaly_magnitude_ts=TimeSeriesData(time=data.time, value=zeros),
            stat_sig_ts=TimeSeriesData(time=data.time, value=zeros),
        )

        return response

    def fit_predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """Fit a model and return the anomaly scores.

        Return AnomalyResponse, when data is passed to it.

        Args:
            data: TimeSeriesData on which detection is run.
            historical_data: Optional; TimeSeriesData corresponding to history. History ends
                exactly where the data begins.

        Returns:
            AnomalyResponse object. The length of this object is same as data. The score property
            gives the score for anomaly.
        """
        self.fit(data=data, historical_data=historical_data)

        return self.predict(data=data, **kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from unittest import TestCase

import pandas as pd

from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers


class DataValidationTest(TestCase):
    def setUp(self) -> None:
        self.TSData = load_air_passengers()

    def test_data_validation(self) -> None:
        # add the extra data point to break the frequency.
        extra_point = pd.DataFrame([["1900-01-01", 2], ["2020-01-01", 2]], columns=["time", "y"])
        DATA = self.TSData.to_dataframe()
        data_with_extra_point = DATA.copy().append(extra_point)

        tsData_with_missing_point = TimeSeriesData(data_with_extra_point)

        tsData_with_missing_point.validate_data(validate_frequency=False, validate_dimension=False)
        tsData_with_missing_point.validate_data(validate_frequency=False, validate_dimension=True)
        with self.assertRaises(ValueError, msg="Frequency validation should fail."):
            tsData_with_missing_point.validate_data(validate_frequency=True, validate_dimension=False)
        with self.assertRaises(ValueError, msg="Frequency validation should fail."):
            tsData_with_missing_point.validate_data(validate_frequency=True, validate_dimension=True)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unittest import TestCase

import numpy as np
import pandas as pd

from kats.consts import TimeSeriesData
from kats.data.utils import load_data
from kats.detectors.seasonality import ACFDetector
from kats.detectors.seasonality import FFTDetector
from kats.models.harmonic_regression import HarmonicRegressionModel


class ACFDetectorTest(TestCase):
    def setUp(self) -> None:
        daily_data = load_data("peyton_manning.csv")
        daily_data.columns = ["time", "y"]
        self.ts_data_daily = TimeSeriesData(daily_data)

        DATA_multi = load_data("multivariate_anomaly_simulated_data.csv")
        self.TSData_multi = TimeSeriesData(DATA_multi)

    def test_acf_detector(self) -> None:
        detector = ACFDetector(data=self.ts_data_daily)
        res = detector.detector(lags=None, diff=1, alpha=0.01)
        self.assertEqual(res["seasonality_presence"], True)
        detector.remover()
        detector.plot()

    def test_no_seasonality(self) -> None:
        np.random.seed(10)
        df_noregress = pd.DataFrame({"no_change": np.random.normal(1, 0.2, 60)})
        df_noregress["time"] = pd.Series(pd.date_range("2019-01-01", "2019-03-01"))
        timeseries = TimeSeriesData(df_noregress)
        detector = ACFDetector(data=timeseries)
        res = detector.detector(lags=None, diff=1, alpha=0.01)
        self.assertEqual(res["seasonality_presence"], False)
        detector.remover()
        detector.plot()

    def test_logging(self) -> None:
        with self.assertRaises(ValueError):
            ACFDetector(data=self.TSData_multi)


class FFTDetectorTest(TestCase):
    def setUp(self) -> None:
        times = pd.to_datetime(np.arange(start=1576195200, stop=1577836801, step=60 * 60), unit="s")
        self.series_times = pd.Series(times)
        harms = HarmonicRegressionModel.fourier_series(self.series_times, 24, 3)
        self.harms_sum = np.sum([1, 1, 1, 1, 1, 1] * harms, axis=1)
        self.data = TimeSeriesData(pd.DataFrame({"time": self.series_times, "values": self.harms_sum}))

        DATA_multi = load_data("multivariate_anomaly_simulated_data.csv")
        self.TSData_multi = TimeSeriesData(DATA_multi)

    def test_detector(self) -> None:
        detector = FFTDetector(data=self.data)
        result = detector.detector()
        self.assertTrue(result["seasonality_presence"])
        self.assertEqual(int(result["seasonalities"][0]), 24)

    def test_logging(self) -> None:
        with self.assertRaises(ValueError):
            FFTDetector(data=self.TSData_multi)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from datetime import datetime
from datetime import timedelta
from unittest import TestCase

import numpy as np
import pandas as pd
from scipy.stats import ks_2samp
from statsmodels.tsa.seasonal import STL
from statsmodels.tsa.seasonal import seasonal_decompose

from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.data.utils import load_data
from kats.detectors.residual_translation import KDEResidualTranslator
from kats.utils.decomposition import TimeSeriesDecomposition
from kats.utils.simulator import Simulator


class DecompositionTest(TestCase):
    def setUp(self) -> None:
        data = load_air_passengers(return_ts=False)
        self.ts_data = TimeSeriesData(data)

        data_nonstandard_name = data.copy()
        data_nonstandard_name.columns = ["ds", "y"]
        self.ts_data_nonstandard_name = TimeSeriesData(df=data_nonstandard_name, time_col_name="ds")

        daily_data = load_data("peyton_manning.csv")
        daily_data.columns = ["time", "y"]
        self.ts_data_daily = TimeSeriesData(daily_data)

        DATA_multi = load_data("multivariate_anomaly_simulated_data.csv")
        self.TSData_multi = TimeSeriesData(DATA_multi)

    def test_asserts(self) -> None:
        with self.assertRaises(ValueError):
            TimeSeriesDecomposition(self.TSData_multi, "additive")

    def test_defaults(self) -> None:
        m1 = TimeSeriesDecomposition(self.ts_data, "additive")
        output1 = m1.decomposer()

        m2 = TimeSeriesDecomposition(self.ts_data, "logarithmic")
        output2 = m2.decomposer()

        self.assertEqual(output1["trend"].value.all(), output2["trend"].value.all())
        self.assertEqual(output1["seasonal"].value.all(), output2["seasonal"].value.all())
        self.assertEqual(output1["rem"].value.all(), output2["rem"].value.all())

        m3 = TimeSeriesDecomposition(self.ts_data, "additive", "STL2")
        output3 = m3.decomposer()

        self.assertEqual(output1["trend"].value.all(), output3["trend"].value.all())
        self.assertEqual(output1["seasonal"].value.all(), output3["seasonal"].value.all())
        self.assertEqual(output1["rem"].value.all(), output3["rem"].value.all())

    def test_nonstandard_time_col_name(self) -> None:
        m = TimeSeriesDecomposition(self.ts_data_nonstandard_name, "multiplicative")
        m.decomposer()
        self.assertEqual(
            # pyre-fixme[16]: `TimeSeriesDecomposition` has no attribute `results`.
            m.results["trend"].time_col_name,
            self.ts_data_nonstandard_name.time_col_name,
        )
        self.assertEqual(
            m.results["seasonal"].time_col_name,
            self.ts_data_nonstandard_name.time_col_name,
        )
        self.assertEqual(m.results["rem"].time_col_name, self.ts_data_nonstandard_name.time_col_name)

    def test_decomposition_additive(self) -> None:
        m = TimeSeriesDecomposition(self.ts_data, "additive")
        output = m.decomposer()

        out = pd.merge(
            pd.DataFrame.from_dict({"time": pd.DatetimeIndex(self.ts_data.time), "y": self.ts_data.value}),
            pd.DataFrame.from_dict(
                {
                    "time": output["trend"].time,
                    "y": output["trend"].value + output["seasonal"].value + output["rem"].value,
                }
            ),
            how="inner",
            on="time",
            suffixes=("_actuals", "_decomposed"),
        )

        self.assertAlmostEqual(np.mean((out["y_actuals"] - out["y_decomposed"]) ** 2), 0, 5)

        m_seasonal = TimeSeriesDecomposition(self.ts_data, "additive", "seasonal_decompose")
        output = m_seasonal.decomposer()

        out = pd.merge(
            pd.DataFrame.from_dict({"time": pd.DatetimeIndex(self.ts_data.time), "y": self.ts_data.value}),
            pd.DataFrame.from_dict(
                {
                    "time": output["trend"].time,
                    "y": output["trend"].value + output["seasonal"].value + output["rem"].value,
                }
            ),
            how="inner",
            on="time",
            suffixes=("_actuals", "_decomposed"),
        )

        self.assertAlmostEqual(np.mean((out["y_actuals"] - out["y_decomposed"]) ** 2), 0, 5)

        m2 = TimeSeriesDecomposition(self.ts_data_daily, "additive")
        output = m2.decomposer()

        out2 = pd.merge(
            pd.DataFrame.from_dict(
                {
                    "time": pd.DatetimeIndex(self.ts_data_daily.time),
                    "y": self.ts_data_daily.value,
                }
            ),
            pd.DataFrame.from_dict(
                {
                    "time": output["trend"].time,
                    "y": output["trend"].value + output["seasonal"].value + output["rem"].value,
                }
            ),
            how="inner",
            on="time",
            suffixes=("_actuals", "_decomposed"),
        )
        self.assertAlmostEqual(np.mean((out2["y_actuals"] - out2["y_decomposed"]) ** 2), 0, 5)

        m2_seasonal = TimeSeriesDecomposition(self.ts_data_daily, "additive", "seasonal_decompose")
        output = m2_seasonal.decomposer()

        out2 = pd.merge(
            pd.DataFrame.from_dict(
                {
                    "time": pd.DatetimeIndex(self.ts_data_daily.time),
                    "y": self.ts_data_daily.value,
                }
            ),
            pd.DataFrame.from_dict(
                {
                    "time": output["trend"].time,
                    "y": output["trend"].value + output["seasonal"].value + output["rem"].value,
                }
            ),
            how="inner",
            on="time",
            suffixes=("_actuals", "_decomposed"),
        )
        self.assertAlmostEqual(np.mean((out2["y_actuals"] - out2["y_decomposed"]) ** 2), 0, 5)

    def test_decomposition_multiplicative(self) -> None:
        m = TimeSeriesDecomposition(self.ts_data, "multiplicative")
        output = m.decomposer()

        out = pd.merge(
            pd.DataFrame.from_dict({"time": pd.DatetimeIndex(self.ts_data.time), "y": self.ts_data.value}),
            pd.DataFrame.from_dict(
                {
                    "time": output["trend"].time,
                    "y": output["trend"].value * output["seasonal"].value * output["rem"].value,
                }
            ),
            how="inner",
            on="time",
            suffixes=("_actuals", "_decomposed"),
        )

        self.assertAlmostEqual(np.mean((out["y_actuals"] - out["y_decomposed"]) ** 2), 0, 5)

        m_seas = TimeSeriesDecomposition(self.ts_data, "multiplicative", "seasonal_decompose")
        output = m_seas.decomposer()

        out = pd.merge(
            pd.DataFrame.from_dict({"time": pd.DatetimeIndex(self.ts_data.time), "y": self.ts_data.value}),
            pd.DataFrame.from_dict(
                {
                    "time": output["trend"].time,
                    "y": output["trend"].value * output["seasonal"].value * output["rem"].value,
                }
            ),
            how="inner",
            on="time",
            suffixes=("_actuals", "_decomposed"),
        )

        self.assertAlmostEqual(np.mean((out["y_actuals"] - out["y_decomposed"]) ** 2), 0, 5)
        m2 = TimeSeriesDecomposition(self.ts_data_daily, "multiplicative")
        output = m2.decomposer()

        out2 = pd.merge(
            pd.DataFrame.from_dict(
                {
                    "time": pd.DatetimeIndex(self.ts_data_daily.time),
                    "y": self.ts_data_daily.value,
                }
            ),
            pd.DataFrame.from_dict(
                {
                    "time": output["trend"].time,
                    "y": output["trend"].value * output["seasonal"].value * output["rem"].value,
                }
            ),
            how="inner",
            on="time",
            suffixes=("_actuals", "_decomposed"),
        )
        self.assertAlmostEqual(np.mean((out2["y_actuals"] - out2["y_decomposed"]) ** 2), 0, 5)

        m2_seas = TimeSeriesDecomposition(self.ts_data_daily, "multiplicative", "seasonal_decompose")
        output = m2_seas.decomposer()

        out2 = pd.merge(
            pd.DataFrame.from_dict(
                {
                    "time": pd.DatetimeIndex(self.ts_data_daily.time),
                    "y": self.ts_data_daily.value,
                }
            ),
            pd.DataFrame.from_dict(
                {
                    "time": output["trend"].time,
                    "y": output["trend"].value * output["seasonal"].value * output["rem"].value,
                }
            ),
            how="inner",
            on="time",
            suffixes=("_actuals", "_decomposed"),
        )
        self.assertAlmostEqual(np.mean((out2["y_actuals"] - out2["y_decomposed"]) ** 2), 0, 5)

    def test_plot(self) -> None:
        m = TimeSeriesDecomposition(self.ts_data, "multiplicative")
        m.decomposer()

        m.plot()

    def test_multiplicative_assert(self) -> None:
        data_new = self.ts_data.to_dataframe().copy()
        data_new["y"] = -1.0 * data_new["y"]
        ts_data_new = TimeSeriesData(data_new)
        print(ts_data_new)
        with self.assertLogs(level="ERROR"):
            m = TimeSeriesDecomposition(ts_data_new, "multiplicative")
            m.decomposer()

    def test_new_freq(self) -> None:
        DATA_multi = self.TSData_multi.to_dataframe()
        df_15_min = DATA_multi[["time", "1"]]
        df_15_min["time"] = list(pd.date_range(end="2020-02-01", periods=df_15_min.shape[0], freq="25T"))
        df_15_min["time"] = df_15_min["time"].astype("str")
        df_15_min.columns = ["time", "y"]

        df_ts = TimeSeriesData(df_15_min)

        m = TimeSeriesDecomposition(df_ts, "additive", method="STL")
        m.decomposer()
        m2 = TimeSeriesDecomposition(df_ts, "additive", method="seasonal_decompose")
        m2.decomposer()

    def test_10_minutes_level_dense_data(self) -> None:
        sim = Simulator(n=2 * 144, freq="10T", start=pd.to_datetime("2021-01-01"))  # 2 days of data
        sim.add_trend(magnitude=1)
        sim.add_seasonality(magnitude=50, period=timedelta(days=1))
        sim.add_noise(magnitude=10)
        # dates are dense, there is no gaps between dates
        dense_dates_ts = sim.stl_sim()
        dense_dates_df = dense_dates_ts.to_dataframe()

        m = TimeSeriesDecomposition(
            dense_dates_ts,
            "additive",
            method="STL",
        )
        decomp = m.decomposer()
        m2 = TimeSeriesDecomposition(
            dense_dates_ts,
            "additive",
            method="STL",
            period=144,
            robust=True,
        )
        decomp2 = m2.decomposer()
        m3 = TimeSeriesDecomposition(
            dense_dates_ts,
            "additive",
            method="seasonal_decompose",
        )
        decomp3 = m3.decomposer()

        # check that interpolate doesn't add new data points
        self.assertEqual(len(decomp["trend"].to_dataframe()), len(dense_dates_ts.to_dataframe()))
        self.assertEqual(len(decomp2["trend"].to_dataframe()), len(dense_dates_ts.to_dataframe()))
        self.assertEqual(len(decomp3["trend"].to_dataframe()), len(dense_dates_ts.to_dataframe()))

        # check if decomposition does what it intends to do
        stl = STL(dense_dates_df.reset_index().value, period=2)
        true_results = stl.fit()
        self.assertTrue((true_results.seasonal.values == decomp["seasonal"].to_dataframe()["season"].values).all())
        self.assertTrue((true_results.trend.values == decomp["trend"].to_dataframe()["trend"].values).all())
        stl = STL(dense_dates_df.reset_index().value, period=144, robust=True)
        true_results = stl.fit()
        self.assertTrue((true_results.seasonal.values == decomp2["seasonal"].to_dataframe()["season"].values).all())
        self.assertTrue((true_results.trend.values == decomp2["trend"].to_dataframe()["trend"].values).all())
        true_results = seasonal_decompose(dense_dates_df.reset_index().value, period=2, model="additive")
        self.assertTrue((true_results.seasonal.values == decomp3["seasonal"].to_dataframe()["seasonal"].values).all())
        self.assertTrue(
            (true_results.trend.values == decomp3["trend"].to_dataframe()["trend"].values)[
                1:-1
            ].all()  # at the beginning and end NaNs
        )

    def test_10_minutes_level_sparse_data(self) -> None:
        # create data
        sim = Simulator(n=2 * 144, freq="10T", start=pd.to_datetime("2021-01-01"))  # 2 days of data
        sim.add_trend(magnitude=1)
        sim.add_seasonality(magnitude=50, period=timedelta(days=1))
        sim.add_noise(magnitude=10)
        dense_dates_ts = sim.stl_sim()
        # dates are sparse, there are some gaps between dates
        sparse_dates_df = dense_dates_ts.to_dataframe().copy()
        sparse_dates_df["time"] = sparse_dates_df["time"].map(
            lambda x: x + pd.Timedelta(365, "D")
            if (x >= pd.Timestamp(2021, 1, 2)) & (x < pd.Timestamp(2021, 1, 3))
            else x
        )
        sparse_dates_ts = TimeSeriesData(sparse_dates_df)

        # do decomposition
        m = TimeSeriesDecomposition(
            sparse_dates_ts,
            "additive",
            method="STL",
        )
        decomp = m.decomposer()
        m2 = TimeSeriesDecomposition(
            sparse_dates_ts,
            "additive",
            method="STL",
            period=144,
            robust=True,
        )
        decomp2 = m2.decomposer()
        m3 = TimeSeriesDecomposition(
            sparse_dates_ts,
            "additive",
            method="seasonal_decompose",
        )
        decomp3 = m3.decomposer()

        # check that interpolate doesn't add new data points
        self.assertEqual(len(decomp["trend"].to_dataframe()), len(dense_dates_ts.to_dataframe()))
        self.assertEqual(len(decomp2["trend"].to_dataframe()), len(dense_dates_ts.to_dataframe()))
        self.assertEqual(len(decomp3["trend"].to_dataframe()), len(dense_dates_ts.to_dataframe()))

        # check if decomposition does what it intends to do
        stl = STL(sparse_dates_df.reset_index().value, period=2)
        true_results = stl.fit()
        self.assertTrue((true_results.seasonal.values == decomp["seasonal"].to_dataframe()["season"].values).all())
        self.assertTrue((true_results.trend.values == decomp["trend"].to_dataframe()["trend"].values).all())
        stl = STL(sparse_dates_df.reset_index().value, period=144, robust=True)
        true_results = stl.fit()
        self.assertTrue((true_results.seasonal.values == decomp2["seasonal"].to_dataframe()["season"].values).all())
        self.assertTrue((true_results.trend.values == decomp2["trend"].to_dataframe()["trend"].values).all())
        true_results = seasonal_decompose(sparse_dates_df.reset_index().value, period=2, model="additive")
        self.assertTrue((true_results.seasonal.values == decomp3["seasonal"].to_dataframe()["seasonal"].values).all())
        self.assertTrue(
            (true_results.trend.values == decomp3["trend"].to_dataframe()["trend"].values)[
                1:-1
            ].all()  # at the beginning and end NaNs
        )


class KDEResidualTranslatorTest(TestCase):
    def setUp(self) -> None:
        data = load_air_passengers(return_ts=False)
        data.columns = ["time", "value"]
        self._y = TimeSeriesData(data)
        yhat = pd.DataFrame({"value": self._y.value.rolling(7).mean().shift(1), "time": self._y.time})
        self._yhat = TimeSeriesData(yhat)
        self._residual = self._y - self._yhat

    def test_setup(self) -> None:
        self.assertEquals(self._yhat.value.isnull().sum(), 7)

    def test_illegal_truncated_fracs(self) -> None:
        with self.assertRaises(ValueError):
            KDEResidualTranslator(-0.1, 0.9)
        with self.assertRaises(ValueError):
            KDEResidualTranslator(1.1, 2.0)
        with self.assertRaises(ValueError):
            KDEResidualTranslator(0.1, -0.9)
        with self.assertRaises(ValueError):
            KDEResidualTranslator(0.1, 1.9)
        with self.assertRaises(ValueError):
            KDEResidualTranslator(0.9, 0.8)

    def test_y_yhat(self) -> None:
        trn = KDEResidualTranslator()
        trn = trn.fit(y=self._y, yhat=self._yhat)
        self._test_residual_trn(trn)

    # def _test_residual(self) -> None:
    #     trn = KDEResidualTranslator()
    #     for name in self._series_names:
    #         dataset = self._get_dataset_for_name(name)[["y", "yhat"]]
    #         dataset["residual"] = dataset.yhat - dataset.y
    #         dataset.drop(["y", "yhat"], axis=1, inplace=True)
    #         trn = trn.fit(dataset)
    #         self._test_residual_trn(trn)

    def _test_residual_trn(self, trn: KDEResidualTranslator) -> None:
        np.testing.assert_allclose(
            np.exp(trn.predict_log_proba(residual=self._residual).value),
            trn.predict_proba(residual=self._residual).value,
        )
        proba = trn.predict_proba(residual=self._residual)
        self.assertTrue(np.all(proba.value >= 0) and np.all(proba.value <= 1))
        ks = ks_2samp(
            # pyre-fixme [16]: Optional type has no attribute `sample`
            trn.kde_.sample(len(self._residual)).flatten(),
            self._residual.value,
        )
        self.assertTrue(ks.statistic < 0.1 or ks.pvalue >= 0.2)


class SimulatorTest(TestCase):
    def test_arima_sim(self) -> None:
        sim = Simulator(n=10, freq="MS", start=pd.to_datetime("2011-01-01 00:00:00"))

        np.random.seed(100)
        ts = sim.arima_sim(ar=[0.1, 0.05], ma=[0.04, 0.1], d=1)

        expected_value = pd.Series(
            [
                0.797342,
                1.494317,
                1.608064,
                1.186103,
                2.147635,
                1.772615,
                0.750320,
                2.159774,
                3.744138,
                3.944730,
            ]
        )
        self.assertEqual(True, (ts.value - expected_value).all())
        self.assertEqual(len(ts.time), 10)

    def test_stl_sim_additive(self) -> None:
        # Create a STL-based simulated object
        sim = Simulator(n=100, freq="1D", start=pd.to_datetime("2011-01-01"))
        np.random.seed(614)
        sim.add_trend(magnitude=10)
        sim.add_seasonality(5, period=timedelta(days=7))
        sim.add_noise(magnitude=2)
        sim_ts = sim.stl_sim()
        # Compare the obtained simulated time series to
        # the original simulated data
        generator1 = Simulator(n=100, freq="D", start="2011-01-01")
        generator1.add_trend(magnitude=10)
        np.random.seed(614)
        generator1.add_seasonality(magnitude=5, period=timedelta(days=7))
        generator1.add_noise(magnitude=2)
        gen_ts_series = generator1.stl_sim()
        # pyre-fixme[16]: `bool` has no attribute `all`.
        self.assertEqual(True, (gen_ts_series.value == sim_ts.value).all())
        self.assertEqual(True, (gen_ts_series.time == sim_ts.time).all())

    def test_stl_sim_multiplicative(self) -> None:
        # Create a STL-based simulated object
        sim = Simulator(n=100, freq="1D", start=pd.to_datetime("2011-01-01"))
        np.random.seed(614)
        sim.add_trend(magnitude=5, multiply=True)
        sim.add_seasonality(10, period=timedelta(days=14))
        sim.add_noise(magnitude=1, multiply=True)
        sim_ts = sim.stl_sim()
        # Compare the obtained simulated time series to
        # the original simulated data
        generator2 = Simulator(n=100, freq="D", start="2011-01-01")
        generator2.add_trend(magnitude=5, multiply=True)
        np.random.seed(614)
        generator2.add_seasonality(magnitude=10, period=timedelta(days=14))
        generator2.add_noise(magnitude=1, multiply=True)
        gen_ts_series = generator2.stl_sim()
        # pyre-fixme[16]: `bool` has no attribute `all`.
        self.assertEqual(True, (gen_ts_series.value == sim_ts.value).all())
        self.assertEqual(True, (gen_ts_series.time == sim_ts.time).all())

    def test_level_shift(self) -> None:
        sim = Simulator(n=450, start="2018-01-01")
        ts = sim.level_shift_sim()

        self.assertEqual(len(ts), 450)

        sim2 = Simulator(n=450, start="2018-01-01")
        ts2 = sim2.level_shift_sim(
            cp_arr=[100, 200],
            level_arr=[3, 20, 2],
            noise=3,
            seasonal_period=7,
            seasonal_magnitude=3,
            anomaly_arr=[50, 150, 250],
            z_score_arr=[10, -10, 20],
        )

        self.assertEqual(len(ts2), 450)

        sim3 = Simulator(n=450, start="2018-01-01")
        ts3 = sim3.level_shift_sim(
            cp_arr=[],
            level_arr=[3],
            noise=3,
            seasonal_period=7,
            seasonal_magnitude=3,
            anomaly_arr=[50, 150, 250],
            z_score_arr=[10, -10, 20],
        )

        self.assertEqual(len(ts3), 450)

    def test_level_shift_mvn_indep(self) -> None:

        sim = Simulator(n=450, start="2018-01-01")
        ts = sim.level_shift_multivariate_indep_sim()
        self.assertEqual(len(ts), 450)

        ts_df = ts.to_dataframe()
        self.assertEqual(ts_df.shape[1], 4)  # time + n_dim

        sim2 = Simulator(n=450, start="2018-01-01")
        ts2 = sim2.level_shift_multivariate_indep_sim(
            cp_arr=[100, 200],
            level_arr=[3, 20, 2],
            noise=3,
            seasonal_period=7,
            seasonal_magnitude=3,
        )

        self.assertEqual(len(ts2), 450)

        ts2_df = ts2.to_dataframe()
        self.assertEqual(ts2_df.shape[1], 4)  # time + n_dim

    def test_trend_shift(self) -> None:
        sim = Simulator(n=450, start="2018-01-01")
        ts = sim.trend_shift_sim()
        self.assertEqual(len(ts), 450)

        sim2 = Simulator(n=450, start="2018-01-01")
        ts2 = sim2.trend_shift_sim(
            cp_arr=[100, 200],
            trend_arr=[3, 20, 2],
            intercept=30,
            noise=30,
            seasonal_period=7,
            seasonal_magnitude=3,
            anomaly_arr=[50, 150, 250],
            z_score_arr=[10, -10, 20],
        )

        self.assertEqual(len(ts2), 450)

        sim3 = Simulator(n=450, start="2018-01-01")
        ts3 = sim3.trend_shift_sim(
            cp_arr=[],
            trend_arr=[3],
            intercept=30,
            noise=30,
            seasonal_period=7,
            seasonal_magnitude=3,
            anomaly_arr=[50, 150, 250],
            z_score_arr=[10, -10, 20],
        )

        self.assertEqual(len(ts3), 450)

    def test_injected_anomalies(self) -> None:
        date_start_str = "2020-03-01"
        date_start = datetime.strptime(date_start_str, "%Y-%m-%d")
        previous_seq = [date_start + timedelta(days=x) for x in range(60)]
        values = np.random.randn(len(previous_seq))
        ts_init = TimeSeriesData(pd.DataFrame({"time": previous_seq[0:60], "y_str": values[0:60]}))

        sim2 = Simulator(n=450, start="2018-01-01")

        ts_level = sim2.inject_level_shift(ts_input=ts_init, cp_arr=[10, 30, 40], level_arr=[10, -10])
        self.assertEqual(len(ts_init), len(ts_level))

        ts_trend = sim2.inject_trend_shift(ts_input=ts_init, cp_arr=[10, 20, 30], trend_arr=[1, -3])
        self.assertEqual(len(ts_init), len(ts_trend))

        ts_spike = sim2.inject_spikes(ts_input=ts_init, anomaly_arr=[10, 20, 30], z_score_arr=[10, -10, 2])
        self.assertEqual(len(ts_init), len(ts_spike))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import warnings
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Sequence
from typing import Union
from typing import cast

try:
    from typing import Protocol
except ImportError:  # pragma: no cover
    from typing_extensions import Protocol  # pragma: no cover

import numpy as np

with warnings.catch_warnings():
    # suppress patsy warning
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from statsmodels.distributions.empirical_distribution import ECDF

# from numpy.typing import ArrayLike
ArrayLike = Union[np.ndarray, Sequence[float]]

# Type aliases
#
# Most metrics have the shape:
#
# def metric(y_true: ArrayLike,
#            y_pred: ArrayLike) -> float:
#
# defined as Metric or
#
# def metric(y_true: ArrayLike,
#            y_pred: ArrayLike,
#            sample_weight: Optional[ArrayLike] = None) -> float:
#
# defined as WeightedMetric.
#
# Other shapes require their own custom protocols.
#
class ArrayMetric(Protocol):
    def __call__(
        self,
        y_true: ArrayLike,
        y_pred: ArrayLike,
        sample_weight: Optional[ArrayLike] = ...,
    ) -> np.ndarray:
        ...  # pragma: no cover


class Metric(Protocol):
    def __call__(self, y_true: ArrayLike, y_pred: ArrayLike) -> float:
        ...  # pragma: no cover


class WeightedMetric(Protocol):
    def __call__(
        self,
        y_true: ArrayLike,
        y_pred: ArrayLike,
        sample_weight: Optional[ArrayLike] = ...,
    ) -> float:
        ...  # pragma: no cover


class MultiOutputMetric(Protocol):
    def __call__(
        self,
        y_true: ArrayLike,
        y_pred: ArrayLike,
        sample_weight: Optional[ArrayLike] = ...,
        multioutput: Union[str, ArrayLike] = ...,
    ) -> float:
        ...  # pragma: no cover


class ThresholdMetric(Protocol):
    def __call__(
        self,
        y_true: ArrayLike,
        y_pred: ArrayLike,
        threshold: float,
    ) -> float:
        ...  # pragma: no cover


class MultiThresholdMetric(Protocol):
    def __call__(
        self,
        y_true: ArrayLike,
        y_pred: ArrayLike,
        threshold: ArrayLike,
    ) -> np.ndarray:
        ...  # pragma: no cover


KatsMetric = Union[
    Metric,
    ArrayMetric,
    WeightedMetric,
    MultiOutputMetric,
    ThresholdMetric,
    MultiThresholdMetric,
]


def _arrays(*args: Optional[ArrayLike]) -> Generator[np.ndarray, None, None]:
    """Ensure all arguments are arrays of matching size.

    Args:
        args: zero or more array-like values.

    Yields:
        The values converted to numpy arrays.

    Raises:
        ValueError if any arrays are different sizes.
    """
    n_samples = None
    for a in args:
        if a is None:
            continue
        if not isinstance(a, np.ndarray):
            a = np.array(a)
        if n_samples is None:
            n_samples = a.shape[0]
        elif a.shape[0] != n_samples:
            raise ValueError(f"Arrays have different number of samples ({a.shape}, expected {n_samples})")
        yield a


def _safe_divide(
    num: np.ndarray,
    denom: Union[np.ndarray, float],
    negative_infinity: float = -1.0,
    positive_infinity: float = 1.0,
    indeterminate: float = 0.0,
    nan: float = np.nan,
) -> np.ndarray:
    """Safely divide one array by another or a float.

    Args:
        num: the numerator
        denom: the denominator
        negative_infinity: the value to replace negative infinity with
        positive_infinity: the value to replace positive infinity with
        indeterminate: the value to replace indeterminates with
        nan: the value to replace nan values with

    Returns:
        The numerator divided by the denominator.
    """
    nans = np.isnan(num)
    if ~np.isscalar(denom):
        nans &= np.isnan(denom)
    with np.errstate(divide="ignore", invalid="ignore"):
        result = np.divide(num, denom)

    # compute locations up-front
    neg_inf = np.isneginf(result)
    pos_inf = np.isposinf(result)
    ind = np.isnan(result)

    # Replace values, nans last
    result[neg_inf] = negative_infinity
    result[pos_inf] = positive_infinity
    result[ind] = indeterminate
    result[nans] = nan
    return result


#
# Array Metrics
#

# Each is defined as a helper on the already-cast values, in case other metrics
# methods need to use those.


def _err(y_true: np.ndarray, y_pred: np.ndarray, sample_weight: Optional[np.ndarray] = None) -> np.ndarray:
    if sample_weight is None:
        return y_true - y_pred
    return _safe_divide(np.multiply(y_true - y_pred, sample_weight), np.nansum(sample_weight))


def error(y_true: ArrayLike, y_pred: ArrayLike, sample_weight: Optional[ArrayLike] = None) -> np.ndarray:
    """Compute the error.

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        sample_weight: optional weights.

    Returns:
        The error array.
    """
    return _err(*_arrays(y_true, y_pred, sample_weight))


def _abs_err(y_true: np.ndarray, y_pred: np.ndarray, sample_weight: Optional[np.ndarray] = None) -> np.ndarray:
    err = np.abs(y_true - y_pred)
    if sample_weight is None:
        return err

    err = np.multiply(err, sample_weight)
    return _safe_divide(err, np.nansum(sample_weight))


def absolute_error(y_true: ArrayLike, y_pred: ArrayLike, sample_weight: Optional[ArrayLike] = None) -> np.ndarray:
    """Compute the absolute error.

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        sample_weight: optional weights.

    Returns:
        The absolute error array.
    """
    return _abs_err(*_arrays(y_true, y_pred, sample_weight))


def _pct_err(y_true: np.ndarray, y_pred: np.ndarray, sample_weight: Optional[np.ndarray] = None) -> np.ndarray:
    err = y_true - y_pred
    if sample_weight is None:
        return _safe_divide(err, y_true)

    err = np.multiply(err, sample_weight)
    return _safe_divide(err, y_true * np.nansum(sample_weight))


def percentage_error(y_true: ArrayLike, y_pred: ArrayLike, sample_weight: Optional[ArrayLike] = None) -> np.ndarray:
    """Compute the percentage error.

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        sample_weight: optional weights.

    Returns:
        The percentage error array.
    """
    return _pct_err(*_arrays(y_true, y_pred, sample_weight))


def _abs_pct_err(y_true: np.ndarray, y_pred: np.ndarray, sample_weight: Optional[np.ndarray] = None) -> np.ndarray:
    err = np.abs(y_true - y_pred)
    if sample_weight is None:
        return _safe_divide(err, y_true)

    err = np.multiply(err, sample_weight)
    return _safe_divide(err, y_true * np.nansum(sample_weight))


def absolute_percentage_error(
    y_true: ArrayLike, y_pred: ArrayLike, sample_weight: Optional[ArrayLike] = None
) -> np.ndarray:
    """Compute the absolute percentage error.

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        sample_weight: optional weights.

    Returns:
        The absolute percentage error array.
    """
    return _abs_pct_err(*_arrays(y_true, y_pred, sample_weight))


#
# Metrics
#


def continuous_rank_probability_score(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the continuous rank in probability score (CRPS).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The linear error in probability score (LEPS) value.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    # If the arrays are empty or all values are missing, return nan.
    if not y_true.size:
        return np.nan
    idx = np.where(~np.isnan(y_pred))[0]
    if not idx.size:
        return np.nan
    y_true = y_true[idx]
    y_pred = y_pred[idx]
    # Otherwise, drop missing entries and compute the ECDF...
    ecdf = ECDF(y_true)
    # ...and finally compute CRPS.
    return np.nanmean((ecdf(y_pred) - ecdf(y_true)) ** 2)


# todo
# def cross_entropy(y_true: ArrayLike,
#                   y_pred: ArrayLike) -> float:


def frequency_exceeds_relative_threshold(y_true: ArrayLike, y_pred: ArrayLike, threshold: float) -> float:
    """Compute the fraction of true that exceeds threshold times prediction.

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        threshold: the threshold value.

    Returns:
        The fraction of true values that exceeds threshold times prediction.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    return np.nanmean(y_true >= threshold * y_pred)


def linear_error_in_probability_space(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the linear error in probability space (LEPS).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The linear error in probability space (LEPS) value.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    # If the arrays are empty or all values are missing, return nan.
    if not y_true.size:
        return np.nan
    idx = np.where(~np.isnan(y_pred))[0]
    if not idx.size:
        return np.nan
    y_true = y_true[idx]
    y_pred = y_pred[idx]
    # Otherwise, drop missing entries and compute the ECDF...
    ecdf = ECDF(y_true)
    # ...and finally compute LEPS.
    return np.nanmean(np.abs(ecdf(y_pred) - ecdf(y_true)))


def median_absolute_error(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the median absolute error (MdAE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The median absolute error (MdAE) value.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    if not y_true.size:
        return np.nan
    return np.nanmedian(_abs_err(y_true, y_pred))


def median_absolute_percentage_error(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the median absolute percentage error (MdAPE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The median absolute percentage error (MdAPE) value.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    if not y_true.size:
        return np.nan
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return np.nanmedian(_abs_pct_err(y_true, y_pred))


def mean_absolute_error(
    y_true: ArrayLike,
    y_pred: ArrayLike,
    sample_weight: Optional[ArrayLike] = None,
    multioutput: Union[str, ArrayLike] = "uniform_average",
) -> float:
    """Compute the mean absolute error (MAE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        sample_weight: optional weights.
        multioutput: raw_values, uniform_avareage, or array shape.

    Returns:
        The mean absolute error (MAE) value.
    """
    if isinstance(multioutput, str):
        if multioutput not in {"raw_values", "uniform_average"}:
            raise ValueError("multioutput must be 'raw_values', 'uniform_average', or an array of floats")

    err = absolute_error(y_true, y_pred)
    if not err.shape[0]:
        return np.nan
    ma = np.ma.MaskedArray(err, np.isnan(err))
    err = np.ma.average(ma, weights=sample_weight, axis=0)
    if isinstance(multioutput, str):
        if multioutput == "raw_values":
            return err
        multioutput = None
    return np.average(err, weights=multioutput)


def mean_absolute_percentage_error(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the mean absolute percentage error (MAPE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The mean absolute percentage error (MAPE) value.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    if not y_true.size:
        return np.nan
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return np.nanmean(_abs_pct_err(y_true, y_pred))


def mean_absolute_scaled_error(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the mean absolute scaled error (MASE).

    MASE is computed by comparing mean absolute error of the predictions to
    the mean absolute error obtained from a lag-1 forecast.

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The mean absolute scaled error (MASE) value.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    if not y_true.size:
        return np.nan
    denom = np.nanmean(np.abs(np.diff(y_true)))
    if np.isclose(denom, 0.0):
        return np.nan
    return mean_absolute_error(y_true, y_pred) / denom


def mean_error(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the mean error.

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The mean error value.
    """
    return np.nanmean(error(y_true, y_pred))


# todo
# def mean_interval_score(y_true: ArrayLike,
#                         y_pred: ArrayLike) -> float:
#


def mean_percentage_error(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the mean percentage error (MPE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The mean percentage error value.
    """
    return np.nanmean(percentage_error(y_true, y_pred))


def mean_squared_error(
    y_true: ArrayLike,
    y_pred: ArrayLike,
    sample_weight: Optional[ArrayLike] = None,
) -> float:
    """Compute the mean squared error (RSE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        sample_weight: optional weights.

    Returns:
        The mean squared error (MSE) value.
    """
    if not len(y_true):
        return np.nan
    if sample_weight is None:
        y_true, y_pred = _arrays(y_true, y_pred)
        return np.nanmean((y_true - y_pred) ** 2)

    y_true, y_pred, sample_weight = _arrays(y_true, y_pred, sample_weight)
    se = np.nanmean((y_true - y_pred) ** 2 * sample_weight, axis=0) / np.average(sample_weight, axis=0)
    return np.nanmean(se)


def root_mean_squared_error(
    y_true: ArrayLike,
    y_pred: ArrayLike,
    sample_weight: Optional[ArrayLike] = None,
) -> float:
    """Compute the root mean squared error (RMSE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        sample_weight: optional weights.

    Returns:
        The root mean squared error (RMSE) value.
    """
    if not len(y_true):
        return np.nan
    return np.sqrt(mean_squared_error(y_true, y_pred, sample_weight))


def root_mean_squared_log_error(
    y_true: ArrayLike,
    y_pred: ArrayLike,
    sample_weight: Optional[ArrayLike] = None,
) -> float:
    """Compute the root mean squared log error (RMSLE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        sample_weight: optional weights.

    Returns:
        The root mean squared log error (RMSLE) value.
    """
    if not len(y_true):
        return np.nan
    return np.sqrt(mean_squared_error(np.log1p(y_true), np.log1p(y_pred), sample_weight))


def root_mean_squared_percentage_error(
    y_true: ArrayLike,
    y_pred: ArrayLike,
    sample_weight: Optional[ArrayLike] = None,
) -> float:
    """Compute the root mean squared percentage error (RMSPE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        sample_weight: optional weights.

    Returns:
        The root mean squared percentage error (RMSPE) value.
    """
    if not len(y_true):
        return np.nan
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return np.sqrt(np.nanmean(_pct_err(*_arrays(y_true, y_pred, sample_weight)) ** 2))


def scaled_symmetric_mean_absolute_percentage_error(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the scaled symmetric mean absolute percentage error (SMAPE).

    Traditionally, SMAPE goes from 0 to 200%. This function goes from
    0 to 100%, and is thus equal to SMAPE divided by 2.

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The symmetric mean absolute percentage error (SMAPE) value.
    """
    num = absolute_error(y_true, y_pred)
    return np.nanmean(_safe_divide(num, np.abs(y_true) + np.abs(y_pred)))


def symmetric_bias(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the symmetric bias (sbias).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The the symmetric bias (sbias) value.
    """
    return -2 * np.nanmean(_safe_divide(error(y_true, y_pred), np.abs(y_true) + np.abs(y_pred)))


def symmetric_mean_absolute_percentage_error(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the symmetric mean absolute percentage error (SMAPE).

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The symmetric mean absolute percentage error (SMAPE) value.
    """
    return 2.0 * scaled_symmetric_mean_absolute_percentage_error(y_true, y_pred)


def tracking_signal(y_true: ArrayLike, y_pred: ArrayLike) -> float:
    """Compute the tracking signal.

    Tracking signal is the ratio of the sum of the residuals to the
    mean absolute error.

    Args:
        y_true: the actual values.
        y_pred: the predicted values.

    Returns:
        The tracking signal value.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    err = mean_absolute_error(y_true, y_pred)
    return np.nan if err == 0 else np.sum(y_true - y_pred) / err


def mult_exceed(y_true: ArrayLike, y_pred: ArrayLike, threshold: ArrayLike) -> np.ndarray:
    """Compute exceed rate for quantile estimates.

    For threshold t (0<t<=0.5), the exceed rate of t is defined as:
        er(y_true, y_pred, t) = mean(y_pred<y_true).
    For threshold t (0.5<t<=1), the exceed rate of t is defined as:
        er(y_true, y_pred, t) = mean(y_pred>y_true).
    For a list threshold T = [t_1, ..., t_d], the exceed rate of T is defined as:
        er(T) = [er(y_true, y_pred, t_1), ..., er(y_true, y_pred, t_d)].

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        threshold: Thresholds to be calculated.

    Returns:
        A `numpy.ndarray` object representing exceed rates.
    """
    if not isinstance(y_true, np.ndarray):
        y_true = np.array(y_true)
    if not isinstance(y_pred, np.ndarray):
        y_pred = np.array(y_pred)
    if not isinstance(threshold, np.ndarray):
        threshold = np.array(threshold)
    if len(y_pred.shape) == 1:
        y_pred = y_pred.reshape(1, -1)
    if len(y_true.shape) == 1:
        y_true = y_true.reshape(1, -1)

    m = len(threshold)
    n, horizon = y_true.shape

    if y_pred.shape[0] != n:
        raise ValueError(f"Arrays have different number of samples ({y_pred.shape}, expected {n, m*horizon})")
    elif y_pred.shape[1] != (m * horizon):
        raise ValueError(f"Arrays have different number of samples ({y_pred.shape}, expected {n, m*horizon})")

    y_true = np.tile(y_true, m)
    mask = np.repeat((threshold > 0.5) * 2 - 1, horizon)

    diff = (y_true - y_pred) * mask > 0
    return np.nanmean(diff.reshape(n, m, -1), axis=2).mean(axis=0)


def pinball_loss(y_true: ArrayLike, y_pred: ArrayLike, threshold: float) -> float:
    """Pinball Loss function module.

    For threshold t (0<t<1), true value y_true and forecast value y_pred, the pinball loss function is defined as:
        For y_true > y_pred,
            pinball(y_true, y_pred, t)=(y_true-y_pred)*t,
        otherwise
            pinball(y_true, y_pred, t)=(y_true-y_pred)*(t-1).

    This module provides functionality for computing pinball loss.

    Attributes:
        y_true: the actual values.
        y_pred: the predicted values.
        threshold:  A float representing the threshold to be calculated.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    if threshold < 0:
        msg = "threshold should not be less than zero."
        logging.error(msg)
        raise ValueError(msg)
    if threshold > 1:
        msg = "threshold should not be greater than one."
        logging.error(msg)
        raise ValueError(msg)

    diff = y_true - y_pred
    return np.nanmean(np.nanmax((diff * threshold, diff * (threshold - 1)), axis=0))


def exceed(y_true: ArrayLike, y_pred: ArrayLike, threshold: float) -> float:
    """Compute exceed rate for quantile estimates.

    For threshold t (0<t<=0.5), the exceed rate of t is defined as:
        er(y_true, y_pred, t) = mean(y_pred<y_true).
    For threshold t (0.5<t<=1), the exceed rate of t is defined as:
        er(y_true, y_pred, t) = mean(y_pred>y_true).
    For a list threshold T = [t_1, ..., t_d], the exceed rate of T is defined as:
        er(T) = [er(y_true, y_pred, t_1), ..., er(y_true, y_pred, t_d)].

    Args:
        y_true: the actual values.
        y_pred: the predicted values.
        threshold: Thresholds to be calculated.

    Returns:
        A `numpy.ndarray` object representing exceed rates.
    """
    y_true, y_pred = _arrays(y_true, y_pred)
    mask = (threshold > 0.5) * 2 - 1
    diff = (y_true - y_pred) * mask > 0
    return np.nanmean(diff)


def coverage(y_true: ArrayLike, y_lower: ArrayLike, y_upper: ArrayLike) -> float:
    """
    Compute the mean coverage rate of the confidence intervals based on the actual values
    Args:
        y_true: the actual values.
        y_lower: the lower bound of the prediction interval.
        y_upper: the upper bound of the prediction interval.

    Returns:
        A float value of the mean coverage rate

    """
    y_true, y_lower, y_upper = _arrays(y_true, y_lower, y_upper)
    diff = (y_lower - y_true <= 0) & (y_true - y_upper <= 0)
    return np.nanmean(diff, axis=0)


def mult_coverage(
    y_true: ArrayLike,
    y_lower: ArrayLike,
    y_upper: ArrayLike,
    rolling_window: Union[None, int] = None,
) -> np.ndarray:
    """
    Compute the coverage rates (or roling mean of the coverage rates) of the confidence intervals based on the actual values
    Args:
        y_true: the actual values.
        y_lower: the lower bound of the prediction interval.
        y_upper: the upper bound of the prediction interval.
        rolling_window: length of the rolling window

    Returns:
        A `numpy.ndarray` object representing coverage rates
    """
    y_true, y_lower, y_upper = _arrays(y_true, y_lower, y_upper)
    diff = (y_lower - y_true <= 0) & (y_true - y_upper <= 0)
    if rolling_window is not None:
        return np.convolve(diff, np.ones(rolling_window), mode="same") / rolling_window

    return diff.astype(int)


def interval_score(y_true: ArrayLike, y_lower: ArrayLike, y_upper: ArrayLike, alpha: float = 0.2) -> float:
    """
    Compute the mean interval score of the confidence intervals based on the actual values
    Args:
        y_true: the actual values.
        y_lower: the lower bound of the prediction interval.
        y_upper: the upper bound of the prediction interval.
        alpha: significance level (1-CI)

    Returns:
        A float value of mean interval score

    """
    y_true, y_lower, y_upper = _arrays(y_true, y_lower, y_upper)
    lower_diff = y_true - y_lower < 0
    upper_diff = y_true - y_upper > 0
    range_ = y_upper - y_lower
    int_score = range_ + lower_diff * ((y_lower - y_true) * 2 / alpha) + upper_diff * ((y_true - y_upper) * 2 / alpha)

    return np.nanmean(int_score, axis=0)


def mult_interval_score(
    y_true: ArrayLike,
    y_lower: ArrayLike,
    y_upper: ArrayLike,
    alpha: float = 0.2,
    rolling_window: Union[None, int] = None,
) -> np.ndarray:
    """
    Compute the interval scores  (or roling mean of the interval scores) of the confidence intervals based on the actual values
    Args:
        y_true: the actual values.
        y_lower: the lower bound of the prediction interval.
        y_upper: the upper bound of the prediction interval.
        alpha: significance level (1-CI)
        rolling_window: length of the rolling window

    Returns:
        A `numpy.ndarray` object representing interval scores

    """
    y_true, y_lower, y_upper = _arrays(y_true, y_lower, y_upper)
    lower_diff = y_true - y_lower < 0
    upper_diff = y_true - y_upper > 0
    range_ = y_upper - y_lower
    int_score = range_ + lower_diff * ((y_lower - y_true) * 2 / alpha) + upper_diff * ((y_true - y_upper) * 2 / alpha)
    if rolling_window is not None:
        return np.convolve(int_score, np.ones(rolling_window), mode="same") / rolling_window

    return int_score


# Name aliases (sorted alphabetically by alias)

ae: ArrayMetric = absolute_error
ape: ArrayMetric = absolute_percentage_error
bias: Metric = mean_error
crps: Metric = continuous_rank_probability_score
leps: Metric = linear_error_in_probability_space
# mad - don't abbreviate this, it's ambiguous whether mean or median
mae: MultiOutputMetric = mean_absolute_error
mape: Metric = mean_absolute_percentage_error
mase: Metric = mean_absolute_scaled_error
mdae: Metric = median_absolute_error
mdape: Metric = median_absolute_percentage_error
me: Metric = mean_error
mean_absolute_deviation: MultiOutputMetric = mean_absolute_error
median_absolute_deviation: Metric = median_absolute_error
# mis: Metric = mean_interval_score  # todo
mpe: Metric = mean_percentage_error
mse: WeightedMetric = mean_squared_error
residual: ArrayMetric = error
rmse: WeightedMetric = root_mean_squared_error
rmsle: WeightedMetric = root_mean_squared_log_error
rmspe: WeightedMetric = root_mean_squared_percentage_error
sbias: Metric = symmetric_bias
scaled_smape: Metric = scaled_symmetric_mean_absolute_percentage_error
smape: Metric = symmetric_mean_absolute_percentage_error
mexceed: MultiThresholdMetric = mult_exceed
ploss: ThresholdMetric = pinball_loss
excd: ThresholdMetric = exceed

ALL_METRICS: Dict[str, KatsMetric] = {
    # Array Metrics
    "error": error,
    "absolute_error": absolute_error,
    "percentage_error": percentage_error,
    "absolute_percentage_error": absolute_percentage_error,
    # Other Metrics
    "mult_exceed": mult_exceed,
    "pinball_loss": pinball_loss,
    "exceed": exceed,
    "continuous_rank_probability_score": continuous_rank_probability_score,
    "frequency_exceeds_relative_threshold": frequency_exceeds_relative_threshold,
    "linear_error_in_probability_space": linear_error_in_probability_space,
    "median_absolute_error": median_absolute_error,
    "median_absolute_percentage_error": median_absolute_percentage_error,
    "mean_absolute_error": mean_absolute_error,
    "mean_absolute_percentage_error": mean_absolute_percentage_error,
    "mean_absolute_scaled_error": mean_absolute_scaled_error,
    "mean_error": mean_error,
    "mean_percentage_error": mean_percentage_error,
    "mean_squared_error": mean_squared_error,
    "root_mean_squared_error": root_mean_squared_error,
    "root_mean_squared_log_error": root_mean_squared_log_error,
    "root_mean_squared_percentage_error": root_mean_squared_percentage_error,
    "symmetric_bias": symmetric_bias,
    "symmetric_mean_absolute_percentage_error": symmetric_mean_absolute_percentage_error,
    "tracking_signal": tracking_signal,
    # Aliases
    "ae": ae,
    "ape": ape,
    "bias": bias,
    "crps": crps,
    "leps": leps,
    "mae": mae,
    "mape": mape,
    "mase": mase,
    "mdae": mdae,
    "mdape": mdape,
    "me": me,
    "mean_absolute_deviation": mean_absolute_deviation,
    "median_absolute_deviation": median_absolute_deviation,
    "mpe": mpe,
    "mse": mse,
    "residual": residual,
    "rmse": rmse,
    "rmsle": rmsle,
    "rmspe": rmspe,
    "sbias": sbias,
    "smape": smape,
}


def metric(name: str) -> KatsMetric:
    """Convenience method to look up a kats.metric by name.

    Args:
        name: the metric name (or abbreviation)

    Returns:
        The metric function.
    """
    try:
        return ALL_METRICS[name]
    except KeyError:
        raise ValueError(f"Could not find metric named {name}")


# These are metrics that return a float and can be invoked with just (y_true, y_pred).
CoreMetric = Union[Metric, WeightedMetric, MultiOutputMetric]

CORE_METRICS: Dict[str, CoreMetric] = {
    name: cast(CoreMetric, method)
    for name, method in ALL_METRICS.items()
    if name
    not in {
        "error",
        "absolute_error",
        "percentage_error",
        "absolute_percentage_error",
        "frequency_exceeds_relative_threshold",
        "ae",
        "ape",
    }
}


def core_metric(name: str) -> CoreMetric:
    """Convenience method to look up a kats.metric by name.

    Args:
        name: the metric name (or abbreviation)

    Returns:
        The metric function.
    """
    try:
        return CORE_METRICS[name]
    except KeyError:
        raise ValueError(f"Could not find core metric named {name}")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

try:
    from importlib import metadata
except ImportError:
    # Python < 3.8
    import importlib_metadata as metadata

from typing import Callable
from typing import Union

from packaging import version as pv

V = Union[str, "Version", pv.Version, pv.LegacyVersion]


class Version:
    """Extend packaging Version to allow comparing to version strings.

    Wraps instead of extends, because pv.parse can return either a
    pv.Version or a pv.LegacyVersion.
    """

    version: Union[pv.Version, pv.LegacyVersion]

    def __init__(self, version: V) -> None:
        """Parse a version.

        Args:
            version: the name of a package, a version string, or a packaging
                version object.
        """
        if isinstance(version, str):
            self.version = self._parse(version)
        elif isinstance(version, Version):
            self.version = version.version
        else:
            self.version = version

    def _parse(self, version: str) -> Union[pv.Version, pv.LegacyVersion]:
        try:
            version = metadata.version(version)
        except metadata.PackageNotFoundError:
            pass
        return pv.parse(version)

    def __lt__(self, other: V) -> bool:
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other: V) -> bool:
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other: V) -> bool:
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other: V) -> bool:
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other: V) -> bool:
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other: V) -> bool:
        return self._compare(other, lambda s, o: s != o)

    def _compare(self, other: V, method: Callable[[V, V], bool]) -> bool:
        if isinstance(other, Version):
            other = other.version
        elif isinstance(other, str):
            other = self._parse(other)
        try:
            return method(self.version._key, other._key)
        except (AttributeError, TypeError):
            return NotImplemented



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
ABDetectorModel conducts an AB test across two concurrent time series. This would be useful
when an experiment is ran between two versions that are logged over time, and there is an
interest when one metric is signficantly different from another.

This implementation supports the following key features:

    1. Rejection Intervals: Sequential rejections are consolidated into contiguous intervals.

    2. Multiple Hypothesis Testing: Multiple Hypothesis Testing occurs when a set of statistical
        inferences occur simultaneously. This is controlled by setting the `duration` parameter
        that only accepts Rejection Intervals of a certain length.

    3. Multiple Distributions: Normal and Binomial likelihoods are available.

    4. Tests of absolute differences (b - a) or relative differences (b / a).

Typical usage example:

>>> # Univariate ABDetectorModel
>>> timeseries = TimeSeriesData(...)
>>> detector = ABDetectorModel()
>>> #Run detector
>>> ab_test_results = detector.fit_predict(data=timeseries)
>>> # Plot the results
>>> detector.plot()
"""

import json
import logging
from dataclasses import dataclass
from enum import Enum
from enum import unique
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from numpy.linalg import matrix_power
from scipy.stats import norm

from kats.consts import IntervalAnomaly
from kats.consts import TimeSeriesData
from kats.detectors.detector import DetectorModel
from kats.detectors.detector_consts import AnomalyResponse
from kats.detectors.detector_consts import ConfidenceBand

DEFAULT_FIGSIZE = (10, 12)


@unique
class ABTestColumnsName(Enum):
    VALUE_A = "value_a"
    VALUE_B = "value_b"
    VARIANCE_A = "variance_a"
    VARIANCE_B = "variance_b"
    SAMPLE_COUNT_A = "sample_count_a"
    SAMPLE_COUNT_B = "sample_count_b"
    EFFECT_SIZE = "effect_size"


NON_NEGATIVE_COLUMNS: List[ABTestColumnsName] = [
    ABTestColumnsName.VARIANCE_A,
    ABTestColumnsName.VARIANCE_B,
]

POSITIVE_COLUMNS: List[ABTestColumnsName] = [
    ABTestColumnsName.SAMPLE_COUNT_A,
    ABTestColumnsName.SAMPLE_COUNT_B,
    ABTestColumnsName.EFFECT_SIZE,
]

INTEGER_COLUMNS: List[ABTestColumnsName] = [
    ABTestColumnsName.SAMPLE_COUNT_A,
    ABTestColumnsName.SAMPLE_COUNT_B,
]

# In the case of Distribution.BINOMIAL, we require that both the
# value_a and value_b columns are proportions (i.e. [0, 1]).
PROPORTION_COLUMNS: List[ABTestColumnsName] = [
    ABTestColumnsName.VALUE_A,
    ABTestColumnsName.VALUE_B,
]


@unique
class ABIntervalType(Enum):
    # Camel case for class names
    FAIL_TO_REJECT = "FailToReject"
    REJECT = "Reject"


@unique
class TestDirection(Enum):
    B_GREATER = "b_greater"
    A_GREATER = "a_greater"


@unique
class Distribution(Enum):
    NORMAL = "normal"
    BINOMIAL = "binomial"


@dataclass
class ABTestResult:
    test_statistic: pd.Series
    stat_sig: pd.Series
    upper: pd.Series
    lower: pd.Series


@unique
class TestStatistic(Enum):
    ABSOLUTE_DIFFERENCE = "absolute_difference"
    RELATIVE_DIFFERENCE = "relative_difference"


class ABInterval(IntervalAnomaly):
    """Extension of IntervalAnomaly that stores ABDetectorModel metadata.

    Used internally to consolidate sequential predictions of the ABDetectorModel
    before returning an AnomalyResponse to the user.
    """

    def __init__(self, interval_type: ABIntervalType, start: pd.Timestamp, end: pd.Timestamp) -> None:
        super().__init__(start=start, end=end)
        self.interval_type: ABIntervalType = interval_type
        self.__start_idx: Optional[int] = None
        self.__end_idx: Optional[int] = None

    @property
    def start_idx(self) -> Optional[int]:
        """Returns the start index wrt to the original TimeSeriesData."""
        return self.__start_idx

    @start_idx.setter
    def start_idx(self, idx: int) -> None:
        if not isinstance(idx, int):
            raise ValueError(f"Expecting start_idx to be of type int. Found {type(idx)}")
        self.__start_idx = idx

    @property
    def end_idx(self) -> Optional[int]:
        """Returns the end index wrt to the original TimeSeriesData."""
        return self.__end_idx

    @end_idx.setter
    def end_idx(self, idx: int) -> None:
        if not isinstance(idx, int):
            raise ValueError(f"Expecting end_idx to be of type int. Found {type(idx)}")
        self.__end_idx = idx

    @property
    def indices(self) -> List[int]:
        """Returns the (inclusive) indices of the interval wrt to the original TimeSeriesData."""
        assert self.start_idx is not None
        assert self.end_idx is not None
        return list(range(self.start_idx, self.end_idx + 1))

    def __repr__(self) -> str:
        _indent = "\n   "
        _repr = f"{self.interval_type.value}Interval("
        _repr += _indent + f"start={self.start},"
        _repr += _indent + f"end={self.end},"
        if self.start_idx is not None and self.end_idx is not None:
            _repr += _indent + f"length={len(self.indices)},"
        _repr += "\n)"
        return _repr


class ABDetectorModel(DetectorModel):
    """ABDetectorModel for detecting whether B - A > effect_size.

    ABDetectorModel runs an AB hypothesis test for each time index.
    Multiple Hypothesis Testing is then mitigated by (optionally) applying a duration
    parameter that consolidates sequential predictions into contiguous intervals.

    Notes:
        - If a duration parameter is not specified, it will be calculated from
        the length of the time series and the overall requested Type I error (alpha).
        - If a duration parameter is specified, just use the supplied duration and alpha
        that is assigned.
        - Only one tailed-tests are currently supported.

    Properties:
        alpha: Overall Type-I error of statistical test. Between 0 and 1 (inclusive).
        corrected_alpha: Corrected Type-I error of the statistical test taking the `duration`
            parameter into consideration. Between 0 and 1 (inclusive).
        duration: length of consecutive predictions considered to be significant.
        test_direction: Test of either b_greater (value_b - value_a > effect_sizes) or
            a_greater (value_a - value_b > effect_sizes).
            Or equivalently, value_b / value_a > (1 + effect_sizes) or
            value_a / value_b > (1 + effect_sizes).
        distribution: The distribution of the data in columns value_a and value_b.
            - For proportions in [0, 1], use Distribution.BINOMIAL
            - For real-valued, use Distribution.NORMAL
        test_statistic: The type of test statistic to compute for each time index.
            - For value_b - value_a, use TestStatistic.ABSOLUTE_DIFFERENCE
            - For value_b / value_a, use TestStatistic.RELATIVE_DIFFERENCE
            Internally, TestStatistic.RELATIVE_DIFFERENCE is tested in log-space and the
            delta method is applied to compute the standard error.
        anomaly_intervals: A list of rejection intervals that meet or exceed the minimal `duration`.
        caution_intervals: Similar to `anomaly_intervals`, but don't meet the minimal `duration`.

    Attributes:
        data: Time series data passed to fit_predict. Must have the schema specified in ABTestColumnsName.
        critical_value: Critical value used to convert test_statistic's to decisions.
        test_result: Results of the AB test.
        fail_to_reject_intervals: List of intervals describing where the test has accepted the null hypothesis.
        reject_intervals: List of intervals describing where the test has rejected the null hypothesis.
    """

    _alpha: Optional[float] = None
    _corrected_alpha: Optional[float] = None
    _duration: Optional[int] = None
    _test_direction: Optional[TestDirection] = None
    _distribution: Optional[Distribution] = None
    _test_statistic: Optional[TestStatistic] = None
    data: Optional[TimeSeriesData] = None
    critical_value: Optional[float] = None
    test_result: Optional[ABTestResult] = None
    fail_to_reject_intervals: Optional[List[ABInterval]] = None
    reject_intervals: Optional[List[ABInterval]] = None

    def __init__(
        self,
        alpha: Optional[float] = 0.05,
        duration: Optional[int] = None,
        serialized_model: Optional[bytes] = None,
        test_direction: Optional[TestDirection] = TestDirection.B_GREATER,
        distribution: Optional[Distribution] = Distribution.BINOMIAL,
        test_statistic: Optional[TestStatistic] = TestStatistic.ABSOLUTE_DIFFERENCE,
    ) -> None:
        if serialized_model:
            model_dict = json.loads(serialized_model)
            self.alpha = model_dict["alpha"]
            self.duration = model_dict["duration"]
            # deserialize value
            self.test_direction = TestDirection(model_dict["test_direction"])
            self.distribution = Distribution(model_dict["distribution"])
            self.test_statistic = TestStatistic(model_dict["test_statistic"])
        else:
            self.alpha = alpha
            self.duration = duration
            self.test_direction = test_direction
            self.distribution = distribution
            self.test_statistic = test_statistic

    def __repr__(self) -> str:
        _indent = "\n   "
        _repr = "ABDetectorModel("
        _repr += _indent + f"alpha={self.alpha},"
        _repr += _indent + f"duration={self.duration},"
        _repr += _indent + f"test_direction={self.test_direction},"
        _repr += _indent + f"test_statistic={self.test_statistic},"
        _repr += "\n)"
        return _repr

    @property
    def alpha(self) -> float:
        if self._alpha is None:
            raise ValueError("alpha is not initialized.")
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: Optional[float]) -> None:
        if alpha is None:
            raise ValueError(f"alpha must be specified. Found {alpha}.")
        # alpha must be specified and between 0 and 1.
        elif alpha < 0 or alpha > 1:
            raise ValueError(f"alpha must be between 0 and 1 (inclusive). Found {alpha}")
        self._alpha = alpha

    @property
    def corrected_alpha(self) -> float:
        if self._corrected_alpha is None:
            raise ValueError("corrected_alpha is not initialized.")
        return self._corrected_alpha

    @corrected_alpha.setter
    def corrected_alpha(self, alpha: Optional[float]) -> None:
        if alpha is None:
            raise ValueError(f"corrected_alpha must be specified. Found {alpha}.")
        # alpha must be specified and between 0 and 1.
        elif alpha < 0 or alpha > 1:
            raise ValueError(f"corrected_alpha must be between 0 and 1 (inclusive). Found {alpha}")
        self._corrected_alpha = alpha

    @property
    def test_direction(self) -> TestDirection:
        if self._test_direction is None:
            raise ValueError("test_direction is not initialized.")
        return self._test_direction

    @test_direction.setter
    def test_direction(self, test_direction: Optional[TestDirection]) -> None:
        if test_direction is None:
            raise ValueError(f"test_direction must be specified. Found {test_direction}.")
        elif not isinstance(test_direction, TestDirection):
            raise TypeError(f"test_direction must be of type TestDirection. Found {type(test_direction)}.")
        self._test_direction = test_direction

    @property
    def distribution(self) -> Distribution:
        if self._distribution is None:
            raise ValueError("distribution is not initialized.")
        return self._distribution

    @distribution.setter
    def distribution(self, distribution: Optional[Distribution]) -> None:
        if distribution is None:
            raise ValueError(f"distribution must be specified. Found {distribution}.")
        elif not isinstance(distribution, Distribution):
            raise TypeError(f"distribution must be of type Distribution. Found {type(distribution)}.")
        self._distribution = distribution

    @property
    def test_statistic(self) -> TestStatistic:
        if self._test_statistic is None:
            raise ValueError("test_statistic is not initialized.")
        return self._test_statistic

    @test_statistic.setter
    def test_statistic(self, test_statistic: Optional[TestStatistic]) -> None:
        if test_statistic is None:
            raise ValueError(f"test_statistic must be specified. Found {test_statistic}.")
        elif not isinstance(test_statistic, TestStatistic):
            raise TypeError(f"test_statistic must be of type TestStatistic. Found {type(test_statistic)}.")
        self._test_statistic = test_statistic

    @property
    def duration(self) -> Optional[int]:
        return self._duration

    @duration.setter
    def duration(self, duration: Optional[int]) -> None:
        # If duration is None, ABDetectorModel will assign a value.
        # Otherwise, it needs to be a positive value.
        if duration is not None and duration <= 0:
            raise ValueError(f"duration must be > 0. Found {duration}.")
        self._duration = duration

    @property
    def anomaly_intervals(self) -> List[ABInterval]:
        _reject_intervals, _duration = self._get_rejection_intervals_and_duration()
        return [interval for interval in _reject_intervals if len(interval.indices) >= _duration]

    @property
    def caution_intervals(self) -> List[ABInterval]:
        _reject_intervals, _duration = self._get_rejection_intervals_and_duration()
        return [interval for interval in _reject_intervals if len(interval.indices) < _duration]

    def _get_rejection_intervals_and_duration(self) -> Tuple[List[ABInterval], int]:
        """Retrieve rejection intervals and minimal duration for post-processing."""
        # reject_intervals and duration must not be None at this point
        if self.reject_intervals is None:
            raise ValueError("reject_intervals are required for anomaly_intervals.")
        _reject_intervals: List[ABInterval] = self.reject_intervals

        if self.duration is None:
            raise ValueError("duration is required for anomaly_intervals.")
        _duration: int = self.duration
        return _reject_intervals, _duration

    def serialize(self) -> bytes:
        """Serializes a model into a json representation."""
        model_dict = {
            "alpha": self.alpha,
            "duration": self.duration,
            # serialize value
            "test_direction": self.test_direction.value,
            "distribution": self.distribution.value,
            "test_statistic": self.test_statistic.value,
        }
        return json.dumps(model_dict).encode("utf-8")

    def fit_predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        consolidate_into_intervals: bool = True,
        interval_padding: int = 30,
        interval_units: str = "m",
        r_tol: float = 0.1,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """Fit and predict on a Interval based AB test on time series data.

        Notes:
            - All entries of data and historical_data must be specified (no na values).
            - All entries of column `effect_size` must be positive (> 0).
            - In the case of Distribution.BINOMIAL, we require that both the
                `value_a` and `value_b` columns are proportions (i.e. [0, 1]).
            - In the case of TestStatistic.RELATIVE_DIFFERENCE, (1 + `effect_size`) will be
                used as the ratio to test value_b / value_a or value_a / value_b.

        Args:
            data: Time series containing columns:
                value_a, value_b, sample_count_a, sample_count_b, effect_size
                and optionally variance_a and variance_b.
            historical_data: Data that will be prepended to `data`.
                Used in an online setting when data is observed data
                in addition to the previous historic observations.
            consolidate_into_intervals: Whether to consolidate sequential predictions
                into contiguous intervals.
            interval_padding: For the case of two sequential anomalies, a ABInterval is well defined.
                However, for a single non-adjacent anomaly, we need to pad the time index to form
                a ABInterval. interval_padding decides how much time to use in this padding.
            interval_units: The units of interval_padding. Includes:
                - D (day)
                - h (hour)
                - m (minute)
                - s (second)
                For less common units and a complete listing, see:
                https://numpy.org/doc/stable/reference/arrays.datetime.html#datetime-units
            r_tol: Relative tolerance used for automatic assignment to duration property.
                If duration is `None`, then a value is automatically assigned such that
                alpha is corrected to be no greater than alpha * (1 + r_tol).

        Returns:
            The results of the Interval based AB test. Including:
                - scores: Raw test statistic.
                - predicted_ts: Boolean array of predictions that are formed from contiguous intervals.
                - stat_sig: Statistical significance of `scores`.
                - upper: Upper limit in the (1 - alpha) confidence interval.
                - lower: Lower limit in the (1 - alpha) confidence interval.
            In the case of TestStatistic.RELATIVE_DIFFERENCE, lower and upper are reported on
                the relative risk scale (exponentiated from log-space).
        """
        self._validate_columns_name(pd.DataFrame(data.value))
        self._validate_data(pd.DataFrame(data.value))
        if historical_data is None:
            self.data = data
        else:
            historical_data.extend(data, validate=False)
            self.data = historical_data
        assert self.data is not None
        _data: TimeSeriesData = self.data
        n = len(_data)
        len_data_zeros = pd.Series(np.zeros(n), copy=False)

        # Step 1: Determine critical_value of test_statistic.
        self.critical_value = self._get_critical_value(length=n, r_tol=r_tol)

        # Step 2: Get a test_statistic for each time value.
        self.test_result = self.get_test_statistic(pd.DataFrame(_data.value))

        # Step 3 (Optional): Consolidate sequential predictions into contiguous intervals.
        #   At this point, we have a computed test statistics for each time index.
        #   To convert these to actual decisions we need to apply:
        #       (1) critical_value
        #       (2) duration
        #   If the user wishes to do this manually, this step can also be bypassed,
        #   returning just the raw test statistics without any critical_value or duration applied.
        if consolidate_into_intervals:
            self.reject_intervals = self._get_intervals(
                time=_data.time,
                interval_type=ABIntervalType.REJECT,
                interval_padding=interval_padding,
                interval_units=interval_units,
            )
            self.fail_to_reject_intervals = self._get_intervals(
                time=_data.time,
                interval_type=ABIntervalType.FAIL_TO_REJECT,
                interval_padding=interval_padding,
                interval_units=interval_units,
            )

        # Step 4: Return to user.
        if self.reject_intervals is not None:
            # If we consolidate_into_intervals, then apply to predicted_ts
            _reject_intervals: List[ABInterval] = self.reject_intervals
            _predicted_ts = self._convert_intervals_to_predictions(time=_data.time, intervals=_reject_intervals)
        else:
            # Otherwise, return a placeholder
            _predicted_ts = TimeSeriesData(time=_data.time, value=len_data_zeros)

        assert self.test_result is not None
        _stat_sig: pd.Series = self.test_result.stat_sig
        _upper: pd.Series = self.test_result.upper
        _lower: pd.Series = self.test_result.lower
        return AnomalyResponse(
            scores=TimeSeriesData(time=_data.time, value=self.test_result.test_statistic),
            confidence_band=ConfidenceBand(
                upper=TimeSeriesData(
                    time=_data.time,
                    value=_upper,
                ),
                lower=TimeSeriesData(
                    time=_data.time,
                    value=_lower,
                ),
            ),
            predicted_ts=_predicted_ts,
            anomaly_magnitude_ts=TimeSeriesData(time=_data.time, value=len_data_zeros),
            stat_sig_ts=TimeSeriesData(time=_data.time, value=_stat_sig),
        )

    def _convert_intervals_to_predictions(self, time: pd.Series, intervals: List[ABInterval]) -> TimeSeriesData:
        # Initialize with all non-predictions
        values = [False] * len(time)
        # At this point, we need to have a duration specified.
        assert self.duration is not None
        _duration: int = self.duration
        # Replace all values for the indices of the ABIntervals
        # if the interval is at least as large as the duration parameter.
        for interval in intervals:
            if len(interval.indices) >= _duration:
                for idx in interval.indices:
                    values[idx] = True
        return TimeSeriesData(time=time, value=pd.Series(values, copy=False))

    @staticmethod
    def _validate_columns_name(df: pd.core.frame.DataFrame) -> None:
        """Determines if required columns are present."""
        for columns in ABTestColumnsName:
            if columns.value not in df.columns:
                raise ValueError(f"{columns.value} must be provided")

    @staticmethod
    def _validate_data(df: pd.core.frame.DataFrame) -> None:
        """Data integrity checks."""
        if df.isnull().values.any():
            raise ValueError("All entries must be specified but na's were found in data.value.")

        for column in POSITIVE_COLUMNS:
            if df[column.value].le(0.0).any():
                raise ValueError(f"{column.value} must be > 0 for each index. Found: \n {df[column.value]}.")

        for column in NON_NEGATIVE_COLUMNS:
            if df[column.value].lt(0.0).any():
                raise ValueError(f"{column.value} must be >= 0 for each index. Found: \n {df[column.value]}.")

        for column in INTEGER_COLUMNS:
            if df[column.value].dtype != np.dtype("int64"):
                raise ValueError(
                    f"{column.value} must be of type int64 for each index. Found: \n {df[column.value].dtype}."
                )

    @staticmethod
    def _validate_binomial_data(df: pd.core.frame.DataFrame) -> None:
        for column in NON_NEGATIVE_COLUMNS:
            if df[column.value].lt(0.0).any() or df[column.value].gt(1.0).any():
                raise ValueError(
                    f"{column.value} must be >= 0.0 and <= 1.0 for each index. Found: \n {df[column.value]}."
                )

    def _get_critical_value(self, length: int, r_tol: float) -> float:
        """Determine a critical value for a statistical test.

        Notes:
            If the user has passed a duration, then they override the automatic assignment.
            In this case, the critical value is directly applied to alpha without consideration of `length`.
            Otherwise, we treat self.alpha initially as the global type-I error, and correct this based
            off `length`, producing an automated assignment for duration and a corrected type-I error.

        Args:
            length: Length of the time series that is being tested.
            r_tol: See `fit_predict()`.

        Returns:
            Critical value used for determing whether to reject or fail to reject the null hypothesis.
        """
        if self.duration is None:
            # Use self.alpha and length to determine best duration
            # and also adjust alpha for the critical_value calculation.
            lowest_m = self._get_lowest_m(p=self.alpha, n=length, r_tol=r_tol)
            logging.warning(
                f"Automatic duration with {length} data points:"
                + f"\nduration set to {lowest_m.m}"
                + f"\ncorrected_alpha set to {lowest_m.p}"
            )
            self.duration = lowest_m.m
            self.corrected_alpha = lowest_m.p
        else:
            # With duration and length fixed, determine an adjusted Type-I error
            # such that the global Type-I error still remains within a
            # relative threshold of self.alpha.
            lowest_p = self._get_lowest_p(m=self.duration, n=length, p_goal=self.alpha, r_tol=r_tol)
            logging.warning(
                f"Type-I Adjustment with {length} data points:"
                + f"\nduration set to {self.duration}"
                + f"\ncorrected_alpha set to {lowest_p.p_corrected}"
                + f"\nType-I Error adjusted to {lowest_p.p_global} from {self.alpha}"
            )
            self.corrected_alpha = lowest_p.p_corrected
        # TODO: Add a condition for a two tailed test. i.e.
        #   if self.two_tailed:
        #       return norm.ppf(1.0 - self.alpha / 2)
        return norm.ppf(1.0 - self.corrected_alpha)

    @dataclass
    class LowestM:
        m: int
        p: float

    def _get_lowest_m(self, p: float, n: int, r_tol: float, max_iter: int = 1000) -> LowestM:
        """Finds lowest m such that the corrected probability is still less than p in n trials.

        Notes:
            This method only bounds the resulting type-I error, it does not gurantee that the
            corrected type-I error is preserved.

        Args:
            p: Overall Type-I error of an experiment.
            n: Total number of trials.
            r_tol: Relative tolerance applied to p.
                i.e. p = 0.5, r_tol = 0.2, p * (r_tol + 1) = 0.6
            max_iter: Total number of iterations before stopping.

        Returns:
            Lowest m and the corrected Type-I error.
        """
        p_global = p
        m = 1
        while m < max_iter and p >= p_global and m <= n:
            # Correct p based off current iter and p_global.
            p = self._probability_of_at_least_one_m_run_in_n_trials(p_global, n=n, m=m)
            if p <= p_global * (r_tol + 1):
                # We have found a solution meeting r_tol.
                return self.LowestM(m=m, p=p)
            # Otherwise, try the next m.
            m += 1
        raise Exception("Automatic duration did not converge. Please explicitly pass duration or revise alpha.")

    @dataclass
    class LowestP:
        p_corrected: float
        p_global: float

    def _get_lowest_p(
        self,
        m: int,
        n: int,
        p_goal: float,
        r_tol: float,
        max_iter: int = 1000,
    ) -> LowestP:
        """Finds a p so that the corrected p is with `r_tol` of `p_global` with n trials and m run size.

        Notes:
            A binary search is performed to find an approximate solution. In cases where this does not
            converge, raising either `max_iter` or `r_tol` will prevent an exception being raised.

        Args:
            m: number of consecutive 1's.
            n: total number of trials.
            p_goal: desired probability of seeing at least one run of m in n trials.
            r_tol: Relative tolerance applied to p_goal, >=1e-9.
                    i.e. p = 0.5, r_tol = 0.2, p * (1 - r_tol) = 0.4 & p * (r_tol + 1) = 0.6
            max_iter: Total number of iterations before stopping.

        Returns:
            A corrected p such that the global p is still within a relative tolerance of `p_goal`.
        """
        # p_goal = p ** n
        if n == m:
            return self.LowestP(p_corrected=p_goal ** (1 / m), p_global=p_goal)

        if r_tol < 1e-9:
            raise ValueError(f"r_tol=1e-9 is the smallest supported value, found: {r_tol}")

        i: int = 0
        # p_goal = U_{i=1}^{n-m} P(run size of m starting at position i)
        #   ≤ ∑_{i=1}^{n-m} P(run size of m starting at position i)
        #   ≤ p ** m * n
        p_low: float = (p_goal / n) ** (1 / m)
        # p_goal ≥ Binomial(k > 0; n//m, p ** m)
        #   = 1 - Binomial(k = 0; n//m, p ** m)
        #   = 1 - (1 - p ** m) ** (n // m)
        p_high: float = (1 - (1 - p_goal) ** (1 / (n // m))) ** (1 / m)
        while p_low <= p_high and i <= max_iter:
            p_corrected = (p_high + p_low) / 2.0
            p_global = self._probability_of_at_least_one_m_run_in_n_trials(p_corrected, n=n, m=m)
            # Return if the corrected Type-I error is within our relative tolerance.
            if p_global <= p_goal * (r_tol + 1) and p_global >= p_goal * (1 - r_tol):
                return self.LowestP(p_corrected=p_corrected, p_global=p_global)
            # Otherwise search higher
            elif p_global < p_goal:
                p_low = p_corrected
            # Otherwise search lower
            elif p_global > p_goal:
                p_high = p_corrected
            i += 1
        raise Exception(
            f"max_iter={max_iter} exceeded while adjusting a goal of {p_goal} with r_tol={r_tol}. Raise max_iter or r_tol."
        )

    @staticmethod
    def _probability_of_at_least_one_m_run_in_n_trials(p: float, n: int, m: int) -> float:
        """P(at least 1 run of m consecutive 1's in n bernoulli trials) in a vectorized formulation.

        Notes:
            Solves the dynamic program using vectorized libraries:

                P(r_j^k) = p^k + Σ_{i=0}^{k-1} p^i * (1 - p) * P(r_{j - i - 1}^k)

                where r_j^k is the event that at least 1 run of m heads occurs in k
                flips with probability p on the jth step.

        Args:
            p: P(x_i = 1) for all i <= n.
            n: total number of trials.
            m: number of consecutive 1's.

        Returns:
            P(at least 1 run of m consecutive 1's in n bernoulli trials).
        """

        def _check_args(p: float, n: int, m: int) -> None:
            if m <= 0:
                raise ValueError(f"m must be > 0. Found m={m}.")
            if n <= 0:
                raise ValueError(f"n must be > 0. Found n={n}.")
            if m > n:
                raise ValueError(f"m must be <= n. Found n={n} and m={m}.")
            if p < 0 or p > 1:
                raise ValueError(f"p must be ∊ (0, 1). Found p={p}.")

        def _vec_solve(p: float, n: int, m: int) -> np.ndarray:
            """Probability of at least 1 run of m consecutive 1's in i bernoulli trials for all i <= n.

            Args:
                p: P(x_i = 1) for all i <= n.
                n: total number of trials.
                m: number of consecutive 1's.

            Returns:
                P(at least 1 run of m consecutive 1's in i trials bernoulli trials) ∀ i<=n
            """
            q = 1 - p

            # state vector => (m + 1,)
            s = np.array([p**m] + [0] * (m - 1) + [1])

            # [p^0 * q, p^1 * q, ..., p^(m - 1) * q, p^m] => (m + 1,)
            f = np.power(p, np.arange(m + 1))
            f *= np.array([q] * m + [1])

            # transition matrix => (m + 1) x (m + 1)
            A = np.diag(v=[1.0] * m, k=1)
            A[:, 0] = f
            A[-2, -1] = 0.0
            A[-1, -1] = 1.0
            return s @ matrix_power(A, n - m)

        # By default, return where i=n from the full state space.
        _check_args(p=p, n=n, m=m)
        return _vec_solve(p=p, n=n, m=m)[0]

    def get_test_statistic(self, df: pd.core.frame.DataFrame) -> ABTestResult:
        if self.distribution == Distribution.BINOMIAL:
            self._validate_binomial_data(df)
            _variance_a = (
                df[ABTestColumnsName.VALUE_A.value]
                * (1 - df[ABTestColumnsName.VALUE_A.value])
                / df[ABTestColumnsName.SAMPLE_COUNT_A.value]
            )
            _variance_b = (
                df[ABTestColumnsName.VALUE_B.value]
                * (1 - df[ABTestColumnsName.VALUE_B.value])
                / df[ABTestColumnsName.SAMPLE_COUNT_B.value]
            )
        elif self.distribution == Distribution.NORMAL:
            _variance_a = df[ABTestColumnsName.VARIANCE_A.value] / df[ABTestColumnsName.SAMPLE_COUNT_A.value]
            _variance_b = df[ABTestColumnsName.VARIANCE_B.value] / df[ABTestColumnsName.SAMPLE_COUNT_B.value]
        else:
            raise ValueError("distribution was incorrectly specified.")
        # Set these internally computed attributes for plotting.
        if self.data is not None:
            self.data.value.variance_a = _variance_a
            self.data.value.variance_b = _variance_b
        return self._get_test_statistic(
            value_a=df[ABTestColumnsName.VALUE_A.value],
            value_b=df[ABTestColumnsName.VALUE_B.value],
            effect_size=df[ABTestColumnsName.EFFECT_SIZE.value],
            variance_a=_variance_a,
            variance_b=_variance_b,
        )

    def _get_test_statistic(
        self,
        value_a: pd.Series,
        value_b: pd.Series,
        effect_size: pd.Series,
        variance_a: pd.Series,
        variance_b: pd.Series,
    ) -> ABTestResult:
        if self.test_direction == TestDirection.B_GREATER:
            _sign: int = 1
        elif self.test_direction == TestDirection.A_GREATER:
            _sign: int = -1
        else:
            raise ValueError(f"test_direction was incorrectly specified. Found {self.test_direction}")

        if self.test_statistic == TestStatistic.ABSOLUTE_DIFFERENCE:
            return self._absolute_difference_test_statistic(
                value_a=value_a,
                value_b=value_b,
                effect_size=effect_size,
                variance_a=variance_a,
                variance_b=variance_b,
                sign=_sign,
            )
        elif self.test_statistic == TestStatistic.RELATIVE_DIFFERENCE:
            return self._relative_difference_test_statistic(
                value_a=value_a,
                value_b=value_b,
                effect_size=effect_size,
                variance_a=variance_a,
                variance_b=variance_b,
                sign=_sign,
            )
        else:
            raise ValueError(f"test_statistic was incorrectly specified. Found {self.test_statistic}")

    def _absolute_difference_test_statistic(
        self,
        value_a: pd.Series,
        value_b: pd.Series,
        effect_size: pd.Series,
        variance_a: pd.Series,
        variance_b: pd.Series,
        sign: int,
    ) -> ABTestResult:
        difference_mean = sign * (value_b - value_a) - effect_size
        difference_std_error = np.sqrt(variance_a + variance_b)
        test_statistic = pd.Series(difference_mean / difference_std_error, copy=False)
        stat_sig = pd.Series(norm.sf(test_statistic), copy=False)

        # TODO: Add a condition for a two tailed test
        assert self.critical_value is not None
        critical_value: float = self.critical_value
        upper = difference_mean + critical_value * difference_std_error
        lower = difference_mean - critical_value * difference_std_error
        return ABTestResult(test_statistic=test_statistic, stat_sig=stat_sig, upper=upper, lower=lower)

    def _relative_difference_test_statistic(
        self,
        value_a: pd.Series,
        value_b: pd.Series,
        effect_size: pd.Series,
        variance_a: pd.Series,
        variance_b: pd.Series,
        sign: int,
    ) -> ABTestResult:
        _EPS = 1e-9
        _EPS_2 = _EPS**2
        # Convert value_a / value_b, consider difference of logs.
        difference_mean = np.log(np.maximum(value_b, _EPS))
        difference_mean -= np.log(np.maximum(value_a, _EPS))
        difference_mean *= sign
        difference_mean -= np.log(1 + effect_size)
        # Apply a delta method for the variance of the log by scaling
        # by a g'(𝜽) ** 2 term. In the case of g = log, g'(𝜽) = 1 / 𝜽.
        # See https://www.stata.com/support/faqs/statistics/delta-method/ for more details.
        difference_std_error = np.sqrt(
            variance_a / np.maximum(value_a**2, _EPS_2) + variance_b / np.maximum(value_b**2, _EPS_2)
        )
        test_statistic = pd.Series(difference_mean / difference_std_error, copy=False)
        stat_sig = pd.Series(norm.sf(test_statistic), copy=False)

        # TODO: Add a condition for a two tailed test
        assert self.critical_value is not None
        critical_value: float = self.critical_value
        upper = np.exp(difference_mean + critical_value * difference_std_error)
        lower = np.exp(difference_mean - critical_value * difference_std_error)
        return ABTestResult(test_statistic=test_statistic, stat_sig=stat_sig, upper=upper, lower=lower)

    def _get_intervals(
        self,
        time: pd.Series,
        interval_type: ABIntervalType,
        interval_padding: int,
        interval_units: str,
    ) -> List[ABInterval]:
        """Consolidates test_result's sequential predictions into contiguous intervals.

        Args:
            time: Time column of the original dataframe.
            interval_type: Name of the interval. One of "accept" or "reject".
            interval_padding: See `fit_predict` for a description.
            interval_units: See `fit_predict` for a description.

        Returns:
            Contiguous intervals made from the sequential predictions of an ABDetectorModel.
        """
        test_result = self.test_result
        if test_result is None:
            raise ValueError("test result is None. Call fit_predict() first")

        # TODO: Add a condition for a two tailed test
        if interval_type == interval_type.REJECT:
            _mask: np.ndarray = (test_result.test_statistic >= self.critical_value).to_numpy()
        elif interval_type == interval_type.FAIL_TO_REJECT:
            _mask: np.ndarray = (test_result.test_statistic < self.critical_value).to_numpy()
        else:
            raise ValueError(f"Expecting test_name one of 'reject' or 'accept'. Found {interval_type.value}.")

        starts, ends = self._get_true_run_indices(_mask)
        intervals: List[ABInterval] = []

        # Loop thru consecutive runs and create Interval objects.
        for start, end in zip(starts, ends):
            start_time = time[start]
            end_time = time[end]
            # Pad the intervals so that single points will have a notion of length.
            start_time -= np.timedelta64(interval_padding, interval_units)
            end_time += np.timedelta64(interval_padding, interval_units)
            interval = ABInterval(interval_type=interval_type, start=start_time, end=end_time)
            # Set attributes of the Interval object. These correspond to the indices
            # of the original TimeSeriesData passed in `fit_predict`.
            interval.start_idx = int(start)
            interval.end_idx = int(end)
            intervals.append(interval)
        return intervals

    @staticmethod
    def _get_true_run_indices(x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Helper function that finds consecutive runs of `True` values.

        Example:
            >>> run_indices = _get_true_run_indices([True, True, False, True])
            >>> print([i.tolist() for i in run_indices])
            >>> [[0, 3], [1, 3]]

        Args:
            x: numpy array of boolean values.

        Returns:
            Tuple of starting & ending index arrays for all consecutive `True` runs in `x`.
        """
        if x.ndim != 1:
            raise ValueError(f"Expecting a 1D array. Found {x.ndim} dimensions.")
        elif x.shape[0] == 0:
            raise ValueError(f"Expecting an array with length > 0. Found {x.shape[0]}.")
        elif x.dtype != np.dtype("bool"):
            raise ValueError(f"x must have x.dtype == np.dtype('bool'). Found {x.dtype}.")
        n = x.shape[0]
        loc_run_start = np.empty(n, dtype=bool)
        loc_run_start[0] = True
        np.not_equal(x[:-1], x[1:], out=loc_run_start[1:])
        run_starts = np.nonzero(loc_run_start)[0]
        run_values = x[loc_run_start]
        run_lengths = np.diff(np.append(run_starts, n))
        return run_starts[run_values], (run_starts + run_lengths - 1)[run_values]

    def predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """Runs `fit_predict`. See `fit_predict` for details."""
        return self.fit_predict(
            data,
            historical_data,
            **kwargs,
        )

    def fit(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> None:
        """Not implemented for this class."""
        NotImplementedError("fit() not implemented. See fit_predict for usage.")

    def plot(
        self,
        figsize: Optional[Tuple[int, int]] = DEFAULT_FIGSIZE,
        interval_units: str = "m",
    ) -> Tuple[plt.Axes, plt.Axes]:
        """Plot the ABTestResult of a ABDetectorModel.

        Warnings:
            This method can only be ran after `fit_predict`.

        Notes:
            To be serializable, the x-axis is plotted in `interval_units` since the start time.

        Args:
            figsize: Figure size.
            interval_units: Units of the intervals. See `fit_predict` for more details.

        Returns:
            Plot of the values & their uncertainty and the test_statistic with overlayed intervals.
        """
        ALPHA = 0.3

        # Data validation
        if self.data is None:
            raise ValueError("Error: data is None. Call fit_predict() before plot()")
        data: Optional[TimeSeriesData] = self.data
        if self.test_result is None:
            raise ValueError("Error: test_result is None. Call fit_predict() before plot()")
        test_result: Optional[ABTestResult] = self.test_result
        assert self.fail_to_reject_intervals is not None
        fail_to_reject_intervals = self.fail_to_reject_intervals

        # Setup axes
        _, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize)

        # X-axis
        x_axis = (data.time - data.time.min()) / np.timedelta64(1, interval_units)

        # Plot the original time series with the respective variance.
        ax1.plot(x_axis, data.value.value_b, label="value_b", ls="-", color="blue")
        ax1.plot(x_axis, data.value.value_a, label="value_a", ls="-", color="teal")
        if self.test_statistic == TestStatistic.ABSOLUTE_DIFFERENCE:
            ax1.plot(
                x_axis,
                data.value.value_a + data.value.effect_size,
                label="threshold",
                ls="--",
                color="red",
            )
        elif self.test_statistic == TestStatistic.RELATIVE_DIFFERENCE:
            ax1.plot(
                x_axis,
                data.value.value_a * (1 + data.value.effect_size),
                label="threshold",
                ls="--",
                color="red",
            )
        a_se = np.sqrt(data.value.variance_a)
        ax1.fill_between(
            x=x_axis,
            y1=data.value.value_a - 1.96 * a_se,
            y2=data.value.value_a + 1.96 * a_se,
            alpha=0.1,
            color="teal",
            label="value_a SE",
        )
        b_se = np.sqrt(data.value.variance_b)
        ax1.fill_between(
            x=x_axis,
            y1=data.value.value_b - 1.96 * b_se,
            y2=data.value.value_b + 1.96 * b_se,
            alpha=0.1,
            color="blue",
            label="value_b SE",
        )
        ax1.set_ylabel("Values")
        ax1.set_xlabel(f"Elapsed time ({interval_units}) from {data.time.min()}")
        ax1.legend()
        ax1.title.set_text("Value for A and B")

        # Plot the test statistic and intervals
        ax2.plot(x_axis, test_result.test_statistic, label="test statistic")
        ax2.axhline(self.critical_value, label="reject", color="r", ls="--")

        # Define two helper functions to extract information from a ABInterval.
        def get_grid(interval: ABInterval) -> pd.Series:
            """Helper function creating a grid of Timestamp.timestamp float values."""
            assert data is not None
            assert interval.end_idx is not None
            end_idx = interval.end_idx
            return (
                pd.concat(
                    [
                        pd.Series(interval.start),
                        pd.Series(data.time[interval.start_idx : end_idx + 1]),
                        pd.Series(interval.end),
                    ]
                )
                - data.time.min()
            ) / np.timedelta64(1, interval_units)

        def get_values(interval: ABInterval) -> pd.Series:
            assert test_result is not None
            assert interval.end_idx is not None
            end_idx = interval.end_idx
            return pd.concat(
                [
                    pd.Series(test_result.test_statistic[interval.start_idx]),
                    pd.Series(test_result.test_statistic[interval.start_idx : end_idx + 1]),
                    pd.Series(test_result.test_statistic[interval.end_idx]),
                ]
            )

        # Plot the fail to reject intervals
        for i, interval in enumerate(fail_to_reject_intervals):
            ax2.fill_between(
                x=get_grid(interval),
                y1=[self.critical_value] * len(get_grid(interval)),
                y2=get_values(interval),
                alpha=ALPHA,
                color="green",
                label="Fail to Reject" if i == 0 else None,
            )

        # Plot the reject intervals less than duration
        for i, interval in enumerate(self.caution_intervals):
            ax2.fill_between(
                x=get_grid(interval),
                y1=[self.critical_value] * len(get_grid(interval)),
                y2=get_values(interval),
                alpha=ALPHA,
                color="yellow",
                label="Reject < Duration" if i == 0 else None,
            )

        # Plot the anomaly intervals more than or equal duration
        for i, interval in enumerate(self.anomaly_intervals):
            ax2.fill_between(
                x=get_grid(interval),
                y1=[self.critical_value] * len(get_grid(interval)),
                y2=get_values(interval),
                alpha=ALPHA,
                color="red",
                label="Reject >= Duration" if i == 0 else None,
            )
        ax2.set_ylabel("Test Statistic")
        ax2.set_xlabel(f"Elapsed time ({interval_units}) from {data.time.min()}")
        ax2.legend()
        ax2.title.set_text(f"Test Statistic for direction: {self.test_direction} & duration: {self.duration}")
        return ax1, ax2



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Forecasting with quadratic model

The quadratic (non-linear) regression model explores a linear relationship
between the forecast variable `y` (observed time series) and predictor
variables `x` and `x^2`, where `x` is the time.
"""

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import numpy as np
import pandas as pd
import statsmodels.api as sm
from statsmodels.sandbox.regression.predstd import wls_prediction_std

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.model import Model


class QuadraticModelParams(Params):
    """Parameter class for Quadratic model.

    This is the parameter class for the quadratic model.
    Attributes:
        alpha: The alpha level for the confidence interval. The default alpha =
        0.05 returns a 95% confidence interval
    """

    def __init__(self, alpha: float = 0.05, **kwargs: Any) -> None:
        super().__init__()
        self.alpha = alpha
        logging.debug("Initialized QuadraticModel parameters. " "alpha:{alpha}".format(alpha=alpha))

    def validate_params(self) -> None:
        """Validate Quadratic Model Parameters

        Since the quadratic model does not require key parameters to be defined
        this is not required for this class
        """
        logging.info("Method validate_params() is not implemented.")
        pass


class QuadraticModel(Model[QuadraticModelParams]):
    """Model class for Quadratic Model.

    This class provides the fit, predict and plot methods for the Quadratic Model

    Attributes:
        data: the input time series data as :class:`kats.consts.TimeSeriesData`
        params: the parameter class defined with `QuadraticModelParams`
    """

    past_length: Optional[int] = None
    model: Optional[Model[QuadraticModelParams]] = None
    freq: Optional[str] = None
    _X_future: Optional[List[int]] = None
    sdev: Optional[float] = None
    dates: Optional[pd.DatetimeIndex] = None
    y_fcst: Optional[pd.Series] = None
    y_fcst_lower: Optional[pd.Series] = None
    y_fcst_upper: Optional[pd.Series] = None
    fcst_df: Optional[pd.DataFrame] = None

    def __init__(self, data: TimeSeriesData, params: QuadraticModelParams) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)

    def fit(self) -> None:
        """fit Quadratic Model."""
        logging.debug("Call fit() with parameters: " "alpha:{alpha}".format(alpha=self.params.alpha))

        # pyre-fixme[16]: `Optional` has no attribute `time`.
        self.past_length = len(self.data.time)
        _X = list(range(self.past_length))
        _X_quad = np.column_stack([_X, np.power(_X, 2)])
        X_quad = sm.add_constant(_X_quad)

        # pyre-fixme[16]: `Optional` has no attribute `value`.
        y = self.data.value
        quad_model = sm.OLS(y, X_quad)

        self.model = quad_model.fit()

    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, include_history: bool = False, **kwargs: Any) -> pd.DataFrame:
        """predict with fitted quadratic model.

        Args:
            steps: the steps or length of the prediction horizon
            include_history: whether to include the historical data in the prediction

        Returns:
            The predicted dataframe with the following columns:
                `time`, `fcst`, `fcst_lower`, and `fcst_upper`
        """
        past_length = self.past_length
        if past_length is None:
            raise ValueError("Call fit() before predict().")
        model = self.model
        assert model is not None
        logging.debug(
            "Call predict() with parameters. " "steps:{steps}, kwargs:{kwargs}".format(steps=steps, kwargs=kwargs)
        )
        # pyre-fixme[16]: `Optional` has no attribute `time`.
        self.freq = kwargs.get("freq", pd.infer_freq(self.data.time))
        self.include_history = include_history

        if include_history:
            _X_future = list(range(0, past_length + steps))
        else:
            _X_future = list(range(past_length, past_length + steps))
        self._X_future = _X_future
        _X_fcst = np.column_stack([_X_future, np.power(_X_future, 2)])
        X_fcst = sm.add_constant(_X_fcst)
        y_fcst = model.predict(X_fcst)
        sdev, y_fcst_lower, y_fcst_upper = wls_prediction_std(self.model, exog=X_fcst, alpha=self.params.alpha)
        self.sdev = sdev
        self.y_fcst = pd.Series(y_fcst, copy=False)
        self.y_fcst_lower = pd.Series(y_fcst_lower, copy=False)
        self.y_fcst_upper = pd.Series(y_fcst_upper, copy=False)

        # create future dates
        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=self.freq)
        self.dates = dates[dates != last_date]

        if include_history:
            self.dates = np.concatenate((pd.to_datetime(self.data.time), self.dates))

        self.fcst_df = fcst_df = pd.DataFrame(
            {
                "time": self.dates,
                "fcst": self.y_fcst,
                "fcst_lower": self.y_fcst_lower,
                "fcst_upper": self.y_fcst_upper,
            },
            copy=False,
        )
        logging.debug("Return forecast data: {fcst_df}".format(fcst_df=self.fcst_df))
        return fcst_df

    def __str__(self) -> str:
        return "Quadratic"

    @staticmethod
    # pyre-fixme[15]: `get_parameter_search_space` overrides method defined in
    #  `Model` inconsistently.
    def get_parameter_search_space() -> List[Dict[str, object]]:
        """get default parameter search space for Quadratic model."""
        return [
            {
                "name": "alpha",
                "type": "choice",
                "value_type": "float",
                "values": [0.01, 0.05, 0.1, 0.25],
                "is_ordered": True,
            },
        ]
